from javastuff import *

class LineStyle:
    STYLE_SOLID = 0
    STYLE_ONOFFDASH = 1
    STYLE_DOUBLEDASH = 2
    STYLE_HIDDEN = 3
    STYLE_INTERNAL = 4

    def __init__(self, id = None, width = 1, color = 0, style = STYLE_SOLID):
        self.id = id
        self.width = width
        self.color = color
        self.style = style

    def printXml(self, file):
        if self.id == 'internal':
            return
        if self.style == self.STYLE_HIDDEN:
            w = -1
            s = self.STYLE_SOLID
        else:
            w = self.width
            s = self.style
        file.write('<Edgestyle id="%s" width="%d" color="%s" style="%d"/>\n' %
                   (self.id, w, toRGB(self.color), s))



#!/usr/bin/env pythonw

import os

resources_subpath = os.path.join('..', 'Resources')
xpngcc_subpath = os.path.join('xpngcc', 'xpngcc.py')
client_subpath = 'XPilotNG.app/Contents/MacOS/XPilotNG'
server_subpath = 'xpilot-ng-server'

my_path = __file__
resources_path = os.path.abspath(os.path.join(os.path.dirname(my_path), resources_subpath))
xpngcc_path = os.path.abspath(os.path.join(resources_path, xpngcc_subpath))
client_path = os.path.abspath(os.path.join(resources_path, client_subpath))
server_path = os.path.abspath(os.path.join(resources_path, server_subpath))

os.environ['xpilot_ng_sdl'] = client_path
os.environ['xpilot_ng_server'] = server_path

os.execl('/usr/bin/env', 'pythonw', xpngcc_path)


from javastuff import *
class MapOptions:
    def __init__(self, opts = None):
        self.opts = {'mapwidth':'3500', 'mapheight':'3500', 'edgewrap':'yes'}
        if opts:
            self.opts.update(opts)
        self.updated()

    def updated(self):
        self.size = Dimension()
        self.size.width = int(self.opts['mapwidth']) * 64
        self.size.height = int(self.opts['mapheight']) * 64
        self.edgeWrap = self.opts['edgewrap'] == 'yes'

    def printXml(self, file):
        file.write('<GeneralOptions>\n')
        # Python 2.2 would have xml.sax.saxutils.quoteattr
        for o in [(x[0], quoteattr(x[1])) for x in self.opts.items()]:
            file.write('<Option name="%s" value="%s"/>\n' % o)
        file.write('</GeneralOptions>\n')



import gtk
import GDK
import copy

from javastuff import *
from config import UNDO_LEVELS

class MouseEvent:
    def __init__(self, point, button, canvas, orig = None):
        self.point = point
        self.x = point.x
        self.y = point.y
        self.button = button
        self.canvas = canvas
        self.orig = orig

class DrawInfo:
    pass

class MapCanvas:
    def __init__(self, area):
        self.eventHandler = None
        self.bmstore = ScaledBitmapStore()
        self.it = None
        self.at = None
        self.posx = 0
        self.posy = 0
        self.area = area
        self.erase = 0
        self.showPoints = 0
        self.filled = 1
        self.textured = 1
        gc = area.get_window().new_gc()
        gc.function = GDK.XOR
        gc.foreground = area.get_colormap().alloc(65535, 65535, 65535)
        gc.line_style = GDK.LINE_ON_OFF_DASH
        self.previewgc = gc
        self.resize()

    def getModel(self):
        return self.model

    def setModel(self, m):
        self.model = m
        self.scale = m.getDefaultScale()
        self.it = None
        self.at = None
        self.initUndo()
        self.repaint()

    def getScale(self):
        return self.scale

    def setScale(self, s):
        self.scale = s
        self.it = None
        self.at = None

    def resize(self):
        widget = self.area
        x, y, w, h = widget.get_allocation()
        self.scratch = gtk.create_pixmap(widget.get_window(), w, h, -1)

    # Set in screen coordinates
    def setCenter(self, x, y):
        p = self.it(Point(x, y))
        self.posx = p.x % self.model.options.size.width
        self.posy = p.y % self.model.options.size.height

    def isErase(self):
        return self.erase

    def setErase(self, erase):
        self.erase = erase

    def setCanvasEventHandler(self, h):
        self.eventHandler = h

    def getCanvasEventHandler(self):
        return self.eventHandler

    def wrapdist2(self, p1, p2):
        mapSize = self.model.options.size
        dx = (p1.x - p2.x) % mapSize.width
        if dx > mapSize.width / 2:
            dx -= mapSize.width
        dy = (p1.y - p2.y) % mapSize.height
        if dy > mapSize.height / 2:
            dy -= mapSize.height
        return float(dx)**2 + float(dy)**2

    def wrap(self, p1, p2):
        s = self.model.options.size
        r= (int(round((p2.x - p1.x + .0) / s.width)) * s.width,
                int(round((p2.y - p1.y + .0) / s.height)) * s.height)
        return r

    def createTransforms(self, width, height):
        self.at = AffineTransform()
        self.at.scale(self.scale, -self.scale)
        self.at.translate(width / self.scale / 2, -height / self.scale / 2)
        self.at.translate(-self.posx, -self.posy)
        self.it = self.at.createInverse()

    def getInverse(self):
        return self.it

    def getTransform(self):
        return self.at

    def wrapOffsets(self, view, b):
        map = self.model.options.size
        nx = (view.x - (b.x + b.width)) / map.width + 1
        xx = (view.x + view.width - b.x) / map.width + 1
        ny = (view.y - (b.y + b.height)) / map.height + 1
        xy = (view.y + view.height - b.y) / map.height + 1
        return [(x * map.width, y * map.height) for x in range(nx, xx)\
                for y in range(ny, xy)]

    def initUndo(self):
        if UNDO_LEVELS > 0:
            self.undos = [None] * (UNDO_LEVELS + 2)
            self.undoIndex = -1
            self.saveUndo()
        else:
            self.undos = None

    def saveUndo(self):
        if not self.undos:
            return
        # The deepcopy can take a visible amount of time, and the X buffer
        # doesn't seem to get completely flushed while in callback from gtk.
        # So explicitly show the results of the user's action so far before
        # doing the copy. There might be a better way to do this (this
        # call apparently actually waits until the drawing is completed by X,
        # which is unnecessary), but I don't know gtk well yet.
        gtk.gdk_flush()
        self.undoIndex += 1
        self.undos[self.undoIndex] = copy.deepcopy(self.model)
        if self.undoIndex > UNDO_LEVELS:
            self.undos = self.undos[1:]
            self.undos.append(None)
            self.undoIndex -= 1

    def undo(self, dir = -1):
        if not self.undos:
            return
        self.undoIndex = max(self.undoIndex + dir, 0)
        model = self.undos[self.undoIndex]
        if model is None:
            self.undoIndex -= 1
        else:
            # Make a copy of the undo state so can return to it again later
            self.model = copy.deepcopy(model)
            self.repaint()

    def repaint(self):
        self.paint(None)

    def paint(self, region):
        area = self.area
        gc = area.get_window().new_gc()
        colormap = area.get_colormap()
        gc.foreground = colormap.alloc(0, 0, 0)
        _, _, aw, ah = area.get_allocation()
        if region:
            x, y, w, h = region
        else:
            w, h = aw, ah
            x = y = 0
        self.createTransforms(aw, ah)
        gtk.draw_rectangle(self.scratch, gc, gtk.TRUE, x, y, w, h)
        mapSize = self.model.options.size
        inv = self.getInverse()
        if region:
            upleft = Point(x, y + h)
            downright = Point(x + w, y)
        else:
            upleft = Point(0, ah)
            downright = Point(aw, 0)
        min = inv(upleft)
        max = inv(downright)
        view = Rectangle(min.x, min.y, max.x - min.x, max.y - min.y)
        self.view = view
        os = self.model.objects[:]
        os.reverse()
        di = DrawInfo()
        di.gc = gc
        di.area = self.scratch
        di.scale = self.scale
        di.colormap = colormap
        di.canvas = self
        di.white_gc = area.get_style().white_gc
        tx2 = self.getTransform()
        for o in os:
            for wrap in self.wrapOffsets(view, o.getBounds()):
                tx = copy.copy(tx2)
                tx.translate(*wrap)
                di.tx = tx
                o.paint(di)
        area.draw_pixmap(area.get_style().fg_gc[gtk.STATE_NORMAL],
                         self.scratch, x, y, x, y, w, h)
        return gtk.TRUE

    def handle(self, area, event):
        if self.model == None:
            return
        if self.eventHandler:
            if event.type == GDK.BUTTON_PRESS:
                evt = MouseEvent(self.it(Point(event.x, event.y)),
                                 event.button, self, event)
                print 'press', evt.point.x, evt.point.y
                self.eventHandler.mousePressed(evt)
            elif event.type == GDK.BUTTON_RELEASE:
                evt = MouseEvent(self.it(Point(event.x, event.y)),
                                 event.button, self, event)
                self.eventHandler.mouseReleased(evt)
            elif event.type == GDK.MOTION_NOTIFY:
                if event.is_hint:
                    x, y = event.window.pointer
                else:
                    x = event.x
                    y = event.y
                evt = MouseEvent(self.it(Point(x, y)), None, self, event)
                self.eventHandler.mouseMoved(evt)
        elif event.type == GDK.BUTTON_PRESS:
            evt = MouseEvent(self.it(Point(event.x, event.y)), event.button,
                             self, event)
            print 'press', evt.point.x, evt.point.y
            for o in self.model.objects:
                if o.checkEvent(self, evt):
                    return

    def drawShape(self, shape, gc = None, target = None):
        if gc is None:
            gc = self.previewgc
        if target is None:
            target = self.area.get_window()
            # Got use use the get_window() because gtk.draw_ methods don't
            # work on a DrawingArea instance, and the paint() methods can't
            # use target.draw_ as they need to work on Pixmaps too and those
            # don't have such methods.
        di = DrawInfo()
        di.gc = gc
        di.area = target
        tx = self.getTransform()
        min = Point()
        max = Point()
        mapSize = self.model.options.size
        for wrap in self.wrapOffsets(self.view, shape.getBounds()):
            di.tx = copy.copy(tx)
            di.tx.translate(*wrap)
            shape.paint(di)



from MapObject import MapObject

# !@# One-team fuels missing, would be easy to implement

class MapFuel(MapObject):
    def __init__(self, x = 0, y = 0, bmstore = None):
        MapObject.__init__(self, x, y, 35 * 64, 35 * 64, bmstore['fuel.gif'])

    def printXml(self, file):
        file.write('<Fuel x="%d" y="%d"/>\n' % (self.x, self.y))



import re
import os
import wx
import wx.lib.mixins.listctrl as listmix
import serverui

class OptionListCtrl(wx.ListCtrl, listmix.ListCtrlAutoWidthMixin):
	def __init__(self, parent, ID, pos=wx.DefaultPosition,
				 size=wx.DefaultSize, style=wx.LC_REPORT|wx.SUNKEN_BORDER):
		wx.ListCtrl.__init__(self, parent, ID, pos, size, style)
		listmix.ListCtrlAutoWidthMixin.__init__(self)
		self.InsertColumn(0, 'Name')
		self.InsertColumn(1, 'Value')
		self.InsertColumn(2, 'Description')
	def set_value(self, row, newval):
		if newval: self.SetStringItem(row, 1, newval)
		else: self.SetStringItem(row, 1, '<default>')
	def set_options(self, opts):
		self.DeleteAllItems()
		row = 0
		for opt in opts:
			self.InsertStringItem(row, opt.names[0])
			if opt.value: self.SetStringItem(row, 1, opt.value)
			else: self.SetStringItem(row, 1, '<default>')
			self.SetStringItem(row, 2, opt.desc)
			row += 1
		if row:
			self.SetColumnWidth(0, wx.LIST_AUTOSIZE)
			self.resizeLastColumn(100)

class OptionViewPanel(wx.Panel):
	def __init__(self, parent):
		wx.Panel.__init__(self, parent, -1)
		self.option = None
		sz = wx.BoxSizer(wx.HORIZONTAL)
		self.SetSizer(sz)
		p = wx.Panel(self)
		ps = wx.FlexGridSizer(2,2)
		ps.AddGrowableCol(1)
		p.SetSizer(ps)
		l = wx.StaticText(p, -1, "Name:")
		l.SetForegroundColour(wx.Color(255,255,255))
		ps.Add(l, 0, wx.ALIGN_CENTER_VERTICAL|wx.ALL, 5)
		self.name = wx.TextCtrl(p, style=wx.TE_READONLY)
		ps.Add(self.name, 0, 
			   wx.EXPAND|wx.ALIGN_CENTER_VERTICAL|wx.ALL, 5)
		l = wx.StaticText(p, -1, "Value:")
		l.SetForegroundColour(wx.Color(255,255,255))
		ps.Add(l, 0, wx.ALIGN_CENTER_VERTICAL|wx.ALL, 5)
		self.value = wx.TextCtrl(p)
		ps.Add(self.value, 0, 
			   wx.EXPAND|wx.ALIGN_CENTER_VERTICAL|wx.ALL, 5)
		sz.Add(p, wx.EXPAND, wx.EXPAND, 0)
		self.desc = wx.TextCtrl(self, -1, 
								style=wx.TE_READONLY|wx.TE_MULTILINE)
		sz.Add(self.desc, wx.EXPAND, wx.EXPAND|wx.ALL, 5)
		p = wx.Panel(self)
		ps = wx.BoxSizer(wx.VERTICAL)
		p.SetSizer(ps)
		b = wx.Button(p, -1, "Apply")
		self.Bind(wx.EVT_BUTTON, self.on_apply, b)
		self.apply = b
		ps.Add(b, 0, wx.BOTTOM, 2)
		b = wx.Button(p, -1, "Default")
		self.Bind(wx.EVT_BUTTON, self.on_default, b)
		self.default = b
		ps.Add(b, 0, wx.TOP, 2)
		sz.Add(p, 0, wx.ALL, 5)
		self.clear()
	def clear(self):
		self.option = None
		self.name.SetValue('')
		self.value.SetValue('')
		self.desc.SetValue('')
		self.apply.Disable()
		self.default.Disable()
	def show_option(self, opt):
		self.apply.Enable(True)
		self.default.Enable(True)
		self.option = opt
		self.name.SetValue(str(opt.names)[1:-1])
		if opt.value: self.value.SetValue(opt.value)
		else: self.value.SetValue('')
		self.desc.SetValue(opt.desc)
	def on_apply(self, evt):
		self.option.value = self.value.GetValue()
		self.GetParent().apply()
	def on_default(self, evt):
		self.option.value = None
		self.value.SetValue('')
		self.GetParent().apply()

class FilterPanel(wx.Panel):
	def __init__(self, parent, action):
		wx.Panel.__init__(self, parent, -1)
		self.action = action
		sz = wx.BoxSizer(wx.HORIZONTAL)
		self.SetSizer(sz)
		l = wx.StaticText(self, -1, "Filter:")
		l.SetForegroundColour(wx.Color(255,255,255))
		sz.Add(l, 0, wx.ALIGN_CENTER_VERTICAL|wx.ALL, 5)
		self.text = wx.TextCtrl(self, -1, '')
		sz.Add(self.text, wx.EXPAND, wx.ALIGN_CENTER_VERTICAL|wx.ALL, 5)
		b = wx.Button(self, -1, "Filter")
		self.Bind(wx.EVT_BUTTON, self.on_filter, b)
		sz.Add(b, 0, wx.ALL, 5)
		b = wx.Button(self, -1, "Show All")
		self.Bind(wx.EVT_BUTTON, self.on_show_all, b)
		sz.Add(b, 0, wx.ALL, 5)
		b = wx.Button(self, -1, action[0])
		self.Bind(wx.EVT_BUTTON, self.on_action, b)
		sz.Add(b, 0, wx.ALL, 5)
	def on_filter(self, evt):
		self.GetParent().filter(self.text.GetValue())
	def on_show_all(self, evt):
		self.text.SetValue('')
		self.GetParent().filter(None)
	def on_action(self, evt):
		self.action[1]()
	def set_filter(self, str):
		if str:	self.text.SetValue(str)

class OptionsPanel(wx.Panel):
	def __init__(self, parent, options, action):
		wx.Panel.__init__(self, parent, -1)
		self.options = options
		self.list = OptionListCtrl(self, -1)
		self.Bind(wx.EVT_LIST_ITEM_SELECTED, self.on_select, self.list)
		sz = wx.BoxSizer(wx.VERTICAL)
		self.filter_panel = FilterPanel(self, action)
		sz.Add(self.filter_panel, 0, wx.EXPAND, 0)
		sz.Add(self.list, wx.EXPAND, wx.EXPAND|wx.ALL, 5)
		self.view = OptionViewPanel(self)
		sz.Add(self.view, 0, wx.EXPAND, 0)
		self.SetSizer(sz)
	def apply(self):
		self.list.set_value(self.row, self.active.value)
	def filter(self, str):
		self.filter_panel.set_filter(str)
		self.view.clear()
		if str:
			self.active_options = [o for o in self.options if o.matches(str) ]
		else: self.active_options = self.options
		self.list.set_options(self.active_options)
	def on_select(self, evt):
		self.row = evt.GetIndex()
		self.active = self.active_options[self.row]
		self.view.show_option(self.active)

class ClientOptionsPanel(OptionsPanel):
	def __init__(self, parent, client, xpilotrc):
		self.xpilotrc = xpilotrc
		opts = parse_options(self, [client, '-help'])
		if os.path.exists(xpilotrc):
			vals = parse_xpilotrc(xpilotrc)
			join_options_with_values(opts, vals)
		sort_options(opts)
		OptionsPanel.__init__(self, parent, opts, ('Save', self.save))
		self.filter(None)
	def save(self):
		if wx.MessageDialog(self, 
							"Save current settings to %s?" % self.xpilotrc,
							"Are you sure?").ShowModal() == wx.ID_OK:
			f = file(self.xpilotrc,'w')
			for o in self.options:
				if o.value:
					o.write(f)
			f.close()

class ServerOptionsPanel(wx.Notebook):
	def __init__(self, parent, client, server, mapdir):
		wx.Notebook.__init__(self, parent, -1, style=wx.BOTTOM|wx.RIGHT)
		self.SetBackgroundColour(wx.Color(0,0,128))
		self.SetForegroundColour(wx.Color(255,255,255))
		self.AddPage(BasicServerOptionsPanel(self, client, server, mapdir),
					 "Basic")
		self.AddPage(AdvancedServerOptionsPanel(self, client, server),
					 "Advanced")
					 
class BasicServerOptionsPanel(wx.Panel):
	def __init__(self, parent, client, server, mapdir):
		wx.Panel.__init__(self, parent, -1)
		self.client = client
		self.server = server
		self.mapdir = mapdir
		sz1 = wx.BoxSizer(wx.HORIZONTAL)
		l = wx.StaticText(self, -1, "Select a map:")
		l.SetForegroundColour(wx.Color(255,255,255))
		sz1.Add((0,0), wx.EXPAND, 0, 0)
		sz1.Add(l, 0, wx.ALIGN_CENTER_VERTICAL|wx.ALL, 5)
		maps = filter(lambda x: x[-4:-1] == '.xp' or x[-3:] == '.xp', 
					  os.listdir(mapdir))
		if maps:
			self.selected = maps[0]
			self.choice = wx.Choice(self, -1, choices=maps)
			self.Bind(wx.EVT_CHOICE, self.on_select, self.choice)
			sz1.Add(self.choice, 0, wx.ALIGN_CENTER_VERTICAL|wx.ALL, 5)
			b = wx.Button(self, -1, "Start")
			self.Bind(wx.EVT_BUTTON, self.start, b)
			sz1.Add(b, 0, wx.ALIGN_CENTER_VERTICAL|wx.ALL, 5)
			sz1.Add((0,0), wx.EXPAND, 0, 0)
			self.SetSizer(sz1)
	def on_select(self, evt):
		self.selected = evt.GetString()
	def start(self, evt):
		opts = [self.server, '-map', os.path.join(self.mapdir, self.selected)]
		console = serverui.Console(self, self.client)
		server = serverui.Server(console, opts)
		console.attach_server(server)
		console.Show()
		server.run()		
		
class AdvancedServerOptionsPanel(OptionsPanel):
	def __init__(self, parent, client, server):
		self.client = client
		self.server = server
		opts = parse_options(self, [server, '-help'])
		sort_options(opts)
		OptionsPanel.__init__(self, parent, opts, ('Start', self.start))
		self.filter('[ Flags: command,')
	def start(self):
		opts = [self.server]
		for o in self.options:
			if o.value:
				if o.type == '<bool>':
					if o.value == 'true':
						opts.append('-' + o.names[0])
					else:
						opts.append('+' + o.names[0])
				else:
					opts.append('-' + o.names[0])
					opts.append(o.value)
		console = serverui.Console(self, self.client)
		server = serverui.Server(console, opts)
		console.attach_server(server)
		console.Show()
		server.run()

class Option:
	def __init__(self, names, type, desc, value):
		self.names = names
		self.type = type
		self.desc = desc
		self.value = value
	def __str__(self):
		return str((self.names, self.type, self.desc, self.value))
	def matches(self, query):
		for name in self.names:
			if name.find(query) != -1: return True
		if self.desc.find(query) != -1: return True
		return False
	def write(self, f):
		f.write('xpilot.' + self.names[0] + ': ' + self.value + '\n')

def parse_options(win, cmd):
	opts = []
	r = re.compile('    -(/\+)?([^<]*)(<\w*>)?')
	p = os.popen(' '.join(cmd))
	try:
		for line in p:
			m = r.match(line)
			if not m: continue
			if m.group(1): type = '<bool>'
			elif m.group(3): type = m.group(3)
			else: continue
			names = [x for x in m.group(2).split() if x != 'or']
			desc = ''
			while 1:
				line = p.next().strip()
				if not line: break
				desc += ' ' + line
			opts.append(Option(names, type, desc.lstrip(), None))
	finally:
		p.close()
	return opts

def parse_xpilotrc(fn):
	vals = {}
	r = re.compile('xpilot\.(\w*)[\s:=]*(.*)')
	f = file(fn)
	try:
		for line in f:
			line = line.strip()
			m = r.match(line)
			if not m: continue
			vals[m.group(1)] = m.group(2)
		return vals
	finally:
		f.close()

def join_options_with_values(opts, vals):
	for opt in opts:
		for name in opt.names:
			if name not in vals:
				continue
			opt.value = vals[name]
			break

def sort_options(opts):
	opts.sort(lambda o1,o2: cmp(o1.names[0], o2.names[0]))



import gtk
from EditorDialog import EditorDialog

class MapObjectPopup:
    def __init__(self, object, event, canvas):
        items = (
            ("/Bring to front", None, self.toFront, 0, None),
            ("/Send to back", None, self.toBack, 0, None),
            ("/Remove", None, self.remove, 0, None),
            ("/Properties", None, self.properties, 0, None)
            )

        # Would be nicer to create the menu only once and then show/hide the
        # same one for each object
        self.object = object
        self.canvas = canvas
        accel_group = gtk.GtkAccelGroup()
        item_factory = gtk.GtkItemFactory(gtk.GtkMenu.get_type(), "<main>",
                                          accel_group)
        item_factory.create_items(items)
        menu = item_factory.get_widget("<main>")
        menu.popup(None, None, None, event.button, event.time)

    def toFront(self, data, menuitem):
        self.canvas.getModel().removeObject(self.object)
        self.canvas.getModel().addToFront(self.object)
        self.canvas.repaint()
        self.canvas.saveUndo()

    def toBack(self, data, menuitem):
        self.canvas.getModel().removeObject(self.object)
        self.canvas.getModel().addToBack(self.object)
        self.canvas.repaint()
        self.canvas.saveUndo()

    def remove(self, data, menuitem):
        self.canvas.getModel().removeObject(self.object)
        self.canvas.repaint()
        self.canvas.saveUndo()

    def properties(self, data, menuitem):
        EditorDialog(self.canvas, self.object)



import gtk
import GDK
import gdkpixbuf
import copy

# This contains classes/functions giving functionality of some Java
# libraries used in the original Java version, and some additional
# utility stuff.

class Rectangle:
    def __init__(self, x, y, w, h):
        self.x = x
        self.y = y
        self.width = w
        self.height = h

    def getBounds(self):
        return self

    def paint(self, di):
        c = copy.copy(self).transform(di.tx)
        gtk.draw_rectangle(di.area, di.gc, gtk.FALSE, c.x, c.y,
                           c.width, c.height)

    def transform(self, tx):
        p1 = tx(Point(self.x, self.y))
        p2 = tx(Point(self.x + self.width, self.y + self.height))
        self.x, self.y = p1.x, p1.y
        self.width = p2.x - p1.x
        self.height = p2.y - p1.y
        if self.width < 0:
            self.x += self.width
            self.width *= -1
        if self.height < 0:
            self.y += self.height
            self.height *= -1
        return self

    def contains(self, p):
        return self.x <= p.x <= self.x + self.width and\
               self.y <= p.y <= self.y + self.height

class Line:
    def __init__(self, p1, p2):
        self.p1 = p1
        self.p2 = p2

    def paint(self, di):
        p1 = di.tx(self.p1)
        p2 = di.tx(self.p2)
        gtk.draw_line(di.area, di.gc, p1.x, p1.y, p2.x, p2.y)

    def getBounds(self):
        p1 = self.p1
        p2 = self.p2
        mx = max(p1.x, p2.x)
        my = max(p1.y, p2.y)
        nx = min(p1.x, p2.x)
        ny = min(p1.y, p2.y)
        return Rectangle(nx, ny, mx - nx, my - ny)

class Point:
    def __init__(self, x = 0, y = 0):
        self.x = int(x)
        self.y = int(y)
        self.width = 0
        self.height = 0

class Polygon:
    def __init__(self):
        self.points = []
        self.npoints = 0

    def recalculate_bounds(self):
        if len(self.points) == 0:
            self.bounds = Rectangle(0, 0, 0, 0)
            return
        x = [p.x for p in self.points]
        y = [p.y for p in self.points]
        self.bounds = Rectangle(min(x), min(y), max(x)-min(x), max(y)-min(y))

    def addPoint(self, x, y):
        self.points.append(Point(x, y))
        self.npoints += 1
        self.recalculate_bounds()

    def getBounds(self):
        return self.bounds

    def transform(self, tx):
        self.points = [tx(p) for p in self.points]
        self.recalculate_bounds()

    def __copy__(self):
        c = Polygon()
        c.points = self.points[:] # Should be just numbers, no deepcopy
        c.npoints = self.npoints
        c.bounds = copy.copy(self.bounds)
        return c

    def contains(self, p):
        l = [Point(p2.x - p.x, p2.y - p.y) for p2 in self.points]
        p1 = l[-1]
        inside = 0
        for p2 in l:
            if p1.y < 0:
                if p2.y >= 0:
                    if p1.x > 0 and p2.x >= 0:
                        inside += 1
                    elif p1.x >= 0 or p2.x >= 0:
                        s = p1.y * (p1.x - p2.x) - p1.x * (p1.y - p2.y)
                        if s == 0:
                            return 1
                        if s > 0:
                            inside += 1
            elif p2.y <= 0:
                if p2.y == 0:
                    if p2.x == 0 or p1.y == 0 and (p1.x <= 0 and p2.x >= 0
                                     or p1.x >= 0 and p2.x <= 0):
                        return 1
                elif p1.x > 0 and p2.x >= 0:
                    inside += 1
                elif p1.x >= 0 or p2.x >= 0:
                    s = p1.y * (p1.x - p2.x) - p1.x * (p1.y - p2.y)
                    if s == 0:
                        return 1
                    if s < 0:
                        inside += 1
            p1 = p2
        return inside & 1


    def paint(self, di):
        gc = di.gc
        tx = di.tx

        trans = [tx(p) for p in self.points]
        points = [(t.x, t.y) for t in trans]
        gtk.draw_polygon(di.area, gc, gtk.FALSE, points)

class AffineTransform:
    def __init__(self, x = 0, y = 0):
        self.x = x
        self.y = y
        self.xscale = 1.
        self.yscale = 1.

    def __call__(self, p):
        return Point(p.x * self.xscale + self.x, p.y * self.yscale + self.y)

    def translate(self, nx, ny):
        self.x += nx * self.xscale
        self.y += ny * self.yscale

    def scale(self, nx, ny):
        self.xscale *= nx
        self.yscale *= ny

    def createInverse(self):
        result = AffineTransform()
        result.xscale = 1. / self.xscale
        result.yscale = 1. / self.yscale
        result.x = -self.x / self.xscale
        result.y = -self.y / self.yscale
        return result

class Dimension:
    pass

#Convert from a single int to a tuple (usable by GTK)
def gcolor(color):
    return (color>>8&65280,color&65280,color<<8&65280)

class MouseEventHandler:
    def mousePressed(self, evt):
        pass

    def mouseReleased(self, evt):
        pass

    def mouseMoved(self, evt):
        pass

def ptSeqDistSq(x1, y1, x2, y2, px, py):
    x2 = float(x2 - x1)
    y2 = float(y2 - y1)
    px = float(px - x1)
    py = float(py - y1)
    s = px * x2 + py * y2
    if s < 0:
        return px * px + py * py
    len = x2 * x2 + y2 * y2
    if s > len:
        return (px - x2)**2 + (py - y2)**2
    s = float(s) / len
    return (px - x2 * s)**2 + (py - y2 * s)**2

class ScaledBitmap:
    def __init__(self, img):
        self.img = img
        self.unscaled, mask = self.img.render_pixmap_and_mask()
        self.scaled = None
        self.scale = 1
    def __deepcopy__(self, memo):
        return self
    def __getitem__(self, scale = 1):
        if scale == 1:
            return self.unscaled
        if scale != self.scale:
            width = int(self.img.width * scale + .5)
            width += width == 0
            height = int(self.img.height * scale + .5)
            height += height == 0
            scaled = self.img.scale_simple(width, height,
                                                gdkpixbuf.INTERP_BILINEAR)
            self.scaled, mask = scaled.render_pixmap_and_mask()
            self.scale = scale
        return self.scaled

import os
import weakref
from config import PIXMAP_PATH
class ScaledBitmapStore:
    def __init__(self):
        self.store = weakref.WeakValueDictionary()
        self.scale = 1
    def __getitem__(self, filename):
        bm = self.store.get(filename)
        if bm is None:
            for path in PIXMAP_PATH:
                if os.path.exists(path + filename):
                    img = gdkpixbuf.new_from_file(path + filename)
                    break
            else:
                raise "Image file not found"
            bm = ScaledBitmap(img)
            self.store[filename] = bm
        return bm

def toRGB(color):
    s = hex(color).upper()[2:] # strip 0x
    s = "0" * (6 - len(s)) + s
    return s

# There would be xml.sax.saxutils.quoteaddr in python 2.2
def quoteattr(text):
    repl = (('&', '&amp;'), ('"', '&quot;'), ("'", '&apos;'),
	    ('<', '&lt;'), ('>', '&gt;'))
    for t, r in repl:
	text = text.replace(t, r)
    return text



from __future__ import nested_scopes
import gtk
import EditorDialog
from constants import polytypes

# The GtkOptionMenu has a rather clumsy interface.
# There might be a nicer widget giving the same functionality in the libraries

class PolygonPropertyEditor(gtk.GtkVBox):
    buttons = EditorDialog.OK_CANCEL
    def __init__(self, canvas, polygon):
        self.polygon = polygon
        self.canvas = canvas
        gtk.GtkVBox.__init__(self)
        self.values = {}
        def makeoptmenu(name, opts, default = 0, func = None):
            hbox = gtk.GtkHBox()
            label = gtk.GtkLabel(name)
            label.show()
            hbox.pack_start(label)
            menu = gtk.GtkMenu()
            def store(widget, data):
                self.values[name] = data
                if func:
                    func(widget, data)
            for text, data in opts:
                menu_item = gtk.GtkMenuItem(text)
                menu_item.connect("activate", store, data)
                menu_item.show()
                menu.append(menu_item)
            option_menu = gtk.GtkOptionMenu()
            option_menu.set_menu(menu)
            option_menu.set_history(default)
            self.values[name] = opts[default][1]
            option_menu.show()
            hbox.pack_start(option_menu)
            hbox.show()
            return hbox
        ps = canvas.model.polyStyles
        _ = [(style.id, style) for style in ps]
        self.pack_start(makeoptmenu("style", _, ps.index(polygon.style)))
        def check_team(widget = None, data = None):
            teammenu.set_sensitive(self.values["type"].has_team)
        _ = [(x.name, x) for x in polytypes]
        self.pack_start(makeoptmenu("type", _, polygon.type, check_team))
        _ = [(str(i), i) for i in range(1, 10)]
        teammenu = makeoptmenu("team", _, polygon.team - 1)
        self.pack_start(teammenu)
        check_team()
        self.show()

    def apply(self):
        self.polygon.style = self.values["style"]
        self.polygon.type = self.values["type"].num
        self.polygon.team = self.values["team"]
        self.canvas.repaint()
        return 1

    def cancel(self):
        return 1



attrs = "name has_team".split()
num = 0
class Polytype:
    def __init__(self, *a):
        global num
        self.num = num
        num += 1
        for name, value in zip(attrs, a):
            setattr(self, name, value)

polytypes = [Polytype(*p) for p in (
    ("Normal", 0),
    ("Ball area", 0),
    ("Ball target", 1),
    ("Decoration", 0),
    )]



#!/usr/bin/env python

import wx
import string
import xputil
import time

class Console(wx.Frame):
	def __init__(self, parent, client):
		wx.Frame.__init__(self, parent, -1, "XPilot NG Server",
						  pos=wx.DefaultPosition,
						  size=(600,300), style=wx.DEFAULT_FRAME_STYLE)
		self.client = client
		bsz = wx.BoxSizer(wx.HORIZONTAL)
		b = wx.Button(self, -1, "Stop")
		b.Disable()
		self.stop = b
		self.Bind(wx.EVT_BUTTON, self.on_close, b)
		bsz.Add(b, 0, wx.ALL, 5)
                if self.client.prog:
		        b = wx.Button(self, -1, "Join")
		        self.Bind(wx.EVT_BUTTON, self.on_join, b)
		        bsz.Add(b, 0, wx.ALL, 5)
		        b.Disable()
		        self.join = b
		sz = wx.BoxSizer(wx.VERTICAL)
		sz.Add(bsz, 0, 0, 0)
		self.area = wx.TextCtrl(self, -1, style=wx.TE_READONLY|wx.TE_MULTILINE)
		sz.Add(self.area, wx.EXPAND, wx.EXPAND|wx.ALL, 5)
		self.SetSizer(sz)
		self.SetAutoLayout(True)
		self.Bind(wx.EVT_CLOSE, self.on_close)
		self.server = None
		self.Bind(wx.EVT_IDLE, self.on_idle)
	def on_close(self, evt):
		self.server.win = None
		if self.server: self.server.kill()
		self.Destroy()
	def on_join(self, evt):
		if self.server: self.client.join("localhost", self.server.port)
	def on_idle(self, evt):
		if self.server:
			stream = self.server.GetInputStream()
			if stream.CanRead():
				self.area.AppendText(stream.readline().decode("iso-8859-1"))
	def attach_server(self, server):
		self.server = server
                if self.client.prog:
		        self.join.Enable(True)
		self.stop.Enable(True)


class Server(xputil.Process):
	def __init__(self, win, argv):
		xputil.Process.__init__(self, win, argv)
		self.port = 15345
		for i in range(len(argv)):
			if argv[i] == '-port':
				self.port = string.atoi(argv[i+1])
				break
	def OnTerminate(self, pid, status):
		if self.win: self.win.Close()
		xputil.Process.OnTerminate(self, pid, status)
		




#!/usr/bin/env python

import sys
import os
import wx
import wx.lib.mixins.listctrl as listmix
import socket
import StringIO
from string import atoi
import config

def fetch(meta):
	s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	try:
		s.connect(meta)
		data = ""
		while True:
			buf = s.recv(1024)
			if not buf:	return parse(data)
			data += buf
	finally:
		s.close()
		
def parse(data):
	lines = StringIO.StringIO(data).readlines()
	servers = []
	for line in lines: servers.append(Server(line.decode('iso-8859-1')))
	return servers

class Server:
	def __init__(self, line):
		self.line = line.strip()
		fields = self.line.split(":")
		self.version = fields[0]
		self.host = fields[1]
		self.port = atoi(fields[2])
		self.count = atoi(fields[3])
		self.mapname = fields[4]
		self.mapsize = fields[5]
		self.author = fields[6]
		self.status = fields[7]
		self.bases = atoi(fields[8])
		self.fps = atoi(fields[9])
		self.playlist = fields[10]
		self.sound = fields[11]
		self.teambases = atoi(fields[13])
		self.uptime = atoi(fields[12])
		self.timing = fields[14]
		self.ip = fields[15]
		self.freebases = fields[16]
		self.queue = atoi(fields[17])
        
	def __str__(self):
		return "Server:" + self.line

class AutoSizeListCtrl(wx.ListCtrl, listmix.ListCtrlAutoWidthMixin):
	def __init__(self, parent, ID, pos=wx.DefaultPosition,
				 size=wx.DefaultSize, style=wx.LC_REPORT|wx.SUNKEN_BORDER):
		wx.ListCtrl.__init__(self, parent, ID, pos, size, style)
		listmix.ListCtrlAutoWidthMixin.__init__(self)

class Panel(wx.Panel):
	def __init__(self, parent, meta, client):
		wx.Panel.__init__(self, parent, -1)
		self.meta = meta
		self.client = client
		self.serverList = AutoSizeListCtrl(self, -1)
		for label in [ "Server", "Version", "#", "Map" ]:
			self.serverList.InsertColumn(sys.maxint, label)
		self.detailList = AutoSizeListCtrl(self, -1)
		for label in [ "Property", "Value" ]:
			self.detailList.InsertColumn(sys.maxint, label)
		for label in [ "Address", "Port", "Size", "Author", "Status", "Bases",
					   "Teams", "Free bases", "Queue", "FPS", "Sound", "Timing",
					   "Uptime"]:
			self.detailList.InsertStringItem(sys.maxint, label)
		self.detailList.Layout()
		self.playerList = AutoSizeListCtrl(self, -1)
		self.playerList.InsertColumn(0, "Players")
		box0 = wx.BoxSizer(wx.HORIZONTAL)
		# FIXME: add a "mute" checkbox to box0
		if (config.client_sdl and config.client_x11):
			self.joinclient = wx.RadioBox(self, label="Client", 
				choices=["Modern", "Classic"])
                        if config.client == config.client_x11:
                                self.joinclient.SetSelection(1)
                        self.client.prog = config.client
			self.Bind(wx.EVT_RADIOBOX, self.OnJoinClient, self.joinclient)
			box0.Add(self.joinclient, 0, wx.ALL)
                        self.mute = wx.CheckBox(self, label="Mute")
                        self.Bind(wx.EVT_CHECKBOX, self.OnMute, self.mute)
			box0.Add((5,0))
			box0.Add(self.mute, 0, wx.ALL)
		elif config.client_sdl:
			self.client.prog = config.client = config.client_sdl
		else:
			self.client.prog = config.client = config.client_x11
		refresh = wx.Button(self, -1, "Refresh")
		self.Bind(wx.EVT_BUTTON, self.OnRefresh, refresh)
		box1 = wx.BoxSizer(wx.HORIZONTAL)
		if config.client:
			join = wx.Button(self, -1, "Join")
			self.Bind(wx.EVT_BUTTON, self.OnJoin, join)
			box1.Add(join, 0, wx.ALL, 5)
			box1.Add((5,0))
		else:
			box1.Add((100,0))
		box1.Add(refresh, 0, wx.ALL, 5)
		box2 = wx.BoxSizer(wx.VERTICAL)
		box2.Add(self.detailList, 2, wx.EXPAND)
		box2.Add(self.playerList, 1, wx.EXPAND)
		box2.Add((1, 5))
		box2.Add(box0)
		box2.Add((1, 5))
		box2.Add(box1)
		box3 = wx.BoxSizer(wx.HORIZONTAL)
		box3.Add(self.serverList, wx.EXPAND, wx.EXPAND|wx.ALL, 5)
		box3.Add(box2, 0, wx.EXPAND|wx.ALL, 5)
		self.SetSizer(box3)
		self.Bind(wx.EVT_LIST_ITEM_SELECTED, self.OnSelect, self.serverList)
	
	def RefreshList(self):
		list = self.serverList
		list.DeleteAllItems()
		self.servers = fetch(self.meta)
		self.servers.sort(lambda x, y: y.count-x.count)
		row = 0
		for server in self.servers:
			list.InsertStringItem(row, server.host)
			list.SetStringItem(row, 1, server.version)
			list.SetStringItem(row, 2, str(server.count))
			list.SetStringItem(row, 3, server.mapname)
			row = row + 1
		for i in range(4):
			list.SetColumnWidth(i, wx.LIST_AUTOSIZE)
		self.serverList.SetItemState(0, wx.LIST_STATE_SELECTED, 
									 wx.LIST_STATE_SELECTED )
		self.Layout()

	def OnJoinClient(self, evt):
		preferredClientString = self.joinclient.GetStringSelection()
		if preferredClientString == 'Modern':
			self.client.prog = config.client = config.client_sdl
		else:
			self.client.prog = config.client = config.client_x11

	def OnMute(self, evt):
		config.is_muted = self.mute.IsChecked()

	def OnSelect(self, evt):
		s = self.servers[evt.GetIndex()]
		self.selected = s
		items = [ s.ip, str(s.port), s.mapsize, s.author, s.status, 
				  str(s.bases), str(s.teambases), str(s.freebases), 
				  str(s.queue), str(s.fps), s.sound, s.timing, 
				  "%.1f days" % (s.uptime/(3600*24.0)) ]
		for i in range(len(items)):
			self.detailList.SetStringItem(i, 1, items[i])
		self.detailList.SetColumnWidth(1, wx.LIST_AUTOSIZE)
		self.playerList.DeleteAllItems()
		if len(s.playlist) > 0:
			for p in s.playlist.split(","):
				self.playerList.InsertStringItem(sys.maxint, p)
			self.playerList.SetColumnWidth(0, wx.LIST_AUTOSIZE)
	
	def OnJoin(self, evt):
		if self.client: 
			self.client.join(self.selected.ip, self.selected.port)
	
	def OnRefresh(self, evt):
		self.RefreshList()
		
class Frame(wx.Frame):
	    def __init__(
			self, parent, ID, title, pos=wx.DefaultPosition,
            size=(800,600), style=wx.DEFAULT_FRAME_STYLE
			):
			wx.Frame.__init__(self, parent, ID, title, pos, size, style)
			panel = Panel(self, ("meta.xpilot.org", 4401), None)
			panel.RefreshList()
			box = wx.BoxSizer(wx.HORIZONTAL)
			box.Add(panel, wx.EXPAND, wx.EXPAND, 0, 0)
			self.SetSizer(box)

class App(wx.App):
	def OnInit(self):
		frame = Frame(None, -1, "XPilot Meta Interface")
		self.SetTopWindow(frame)
		frame.Show(True)
		return True

def main():
	app = App(0)
	app.MainLoop()

if __name__ == '__main__':
	main()



from __future__ import nested_scopes # required in Python 2.1
import gtk

OK_CANCEL = 0
CLOSE = 1

class EditorDialog(gtk.GtkDialog):
    def __init__(self, canvas, object):
        gtk.GtkDialog.__init__(self)
        editor = object.getPropertyEditor(canvas)
        self.set_title("Property editor")
        self.vbox.pack_start(editor)
        def addbutton(text, function):
            b = gtk.GtkButton(text)
            b.connect("clicked", function)
            self.action_area.pack_start(b)
            b.show()
        def ok(button):
            if not editor.apply():
                return
            self.hide()
            self.destroy()
        def cancel(button):
            if not editor.cancel():
                return
            self.hide()
            self.destroy()
        if editor.buttons == OK_CANCEL:
            addbutton("OK", ok)
            addbutton("Cancel", cancel)
        else:
            addbutton("Close", ok)
        # gtk.grab_add(self) # modal
        self.show()



from javastuff import *
from MapObject import MapObject
from PolygonStyle import PolygonStyle
from LineStyle import LineStyle
from PolygonPropertyEditor import PolygonPropertyEditor

import gtk
import GDK

class MapPolygon(MapObject):
    TYPE_NORMAL = 0
    TYPE_BALLAREA = 1
    TYPE_BALLTARGET = 2
    TYPE_DECORATION = 3

    def __init__(self, p = None, style = None, edgeStyles = None,bmstore=None):
        self.polygon = p
        self.style = style
        self.edgeStyles = edgeStyles
        self.type = self.TYPE_NORMAL
        self.team = 1

    def getCreateHandler(self):
        return CreateHandler(self)

    def getBounds(self):
        return self.polygon.getBounds()

    def contains(self, canvas, p):
        for wrap in canvas.wrapOffsets(self.getBounds(), p):
            if self.polygon.contains(Point(p.x + wrap[0], p.y + wrap[1])):
                return 1
        return 0

    def getZOrder(self):
        return 20

    def getPreviewShape(self):
        return self.polygon

    def moveTo(self, x, y):
        self.polygon.transform(AffineTransform(x - self.x, y - self.y))

    def paint(self, di):
        gc = di.gc
        colormap = di.colormap
        tx = di.tx

        trans = [tx(p) for p in self.polygon.points]

        bounds = self.polygon.getBounds()
        if max(bounds.width, bounds.height) * di.canvas.scale > 30000:
            # gtk doesn't like getting coordinates bigger than 32767.
            # Those cause an illogical error message
            # "TypeError: sequence member not a 2-tuple"
            for p in trans:
                if p.x > 30000:
                    p.x = 30000
                elif p.x < -30000:
                    p.x = -30000
                if p.y > 30000:
                    p.y = 30000
                elif p.y < -30000:
                    p.y = -30000

        points = [(t.x, t.y) for t in trans]

        nofill = 1
        if self.style.textured and di.canvas.textured:
            if self.style.texture.scalable:
                gc.tile = self.style.texture.image[di.scale * 64]
            else:
                gc.tile = self.style.texture.image[1]
            # !@# Should align the texture
            gc.fill = GDK.TILED
            nofill = 0
        elif self.style.filled and di.canvas.filled:
            gc.foreground = colormap.alloc(*gcolor(self.style.color))
            nofill = 0
        if not nofill:
            gtk.draw_polygon(di.area, gc, gtk.TRUE, points)
            gc.fill = GDK.SOLID

        if self.edgeStyles is None:
            df = self.style.defaultEdgeStyle
            if df.style != LineStyle.STYLE_HIDDEN or nofill:
                gc.foreground = colormap.alloc(*gcolor(df.color))
                gtk.draw_polygon(di.area, gc, gtk.FALSE, points)
                # !@# dashed lines, width missing here
        else:
            si = 0
            cs = self.edgeStyles[0]
            for i in range(1, len(points)):
                if self.edgeStyles[i] == cs:
                    continue
                lstyle = cs or self.style.defaultEdgeStyle
                if lstyle.style != LineStyle.STYLE_HIDDEN or nofill:
                    gc.foreground = colormap.alloc(*gcolor(lstyle.color))
                    gtk.draw_lines(di.area, gc, points[si:i+1])
                si = i
                cs = self.edgeStyles[i]
            points2 = points[si:]
            points2.append(points[0])
            lstyle = cs or self.style.defaultEdgeStyle
            if lstyle.style != LineStyle.STYLE_HIDDEN or nofill:
                gc.foreground = colormap.alloc(*gcolor(lstyle.color))
                gtk.draw_lines(di.area, gc, points2)

        if di.canvas.showPoints:
            for p in points:
                gtk.draw_rectangle(di.area, di.white_gc, gtk.FALSE,
                                   p[0] - 1, p[1] - 1, 2, 2)

    def getPropertyEditor(self, canvas):
        return PolygonPropertyEditor(canvas, self)

    def checkEvent(self, canvas, me):
        b = self.getBounds()
        p = me.point
        pl = self.polygon
        THRESHOLD = 9 / canvas.scale**2

        larger = Rectangle(b.x - 20 * 64, b.y - 20 * 64,
                           b.width + 40 * 64, b.height + 40 * 64)
        wraps = canvas.wrapOffsets(larger, p)
        if len(wraps) == 0:
            return 0

        if me.button == 1:
            for i in range(len(pl.points)):
                point = pl.points[i]
                if canvas.wrapdist2(p, point) < THRESHOLD:
                    if canvas.isErase():
                        if self.polygon.npoints > 3:
                            self.removePoint(i)
                            canvas.repaint()
                            canvas.saveUndo()
                    else:
                        canvas.setCanvasEventHandler(PolygonPointMoveHandler(
                            self, me, i, canvas.wrap(p, point)))
                    return 1

        if not canvas.isErase():
            for wrap in wraps:
                for i in range(pl.npoints):
                    p1 = pl.points[i - 1]
                    p2 = pl.points[i]
                    if ptSeqDistSq(p1.x, p1.y, p2.x, p2.y, p.x + wrap[0],
                                   p.y + wrap[1]) < THRESHOLD:
                        if me.button == 1:
                            self.insertPoint(i, p)
                            canvas.setCanvasEventHandler(PolygonPointMoveHandler(
                                self, me, i, wrap))
                            return 1

        # MapObject's move handler needs x and y. Would be nicer to trap
        # their access and return appropriate values then...
        self.x = self.polygon.points[0].x
        self.y = self.polygon.points[0].y
        return MapObject.checkEvent(self, canvas, me)

    def removePoint(self, i):
        del self.polygon.points[i]
        self.polygon.npoints -= 1
        self.polygon.recalculate_bounds()
        if self.edgeStyles is not None:
            del self.edgeStyles[i - 1] # works right for i == 0

    def insertPoint(self, i, p):
        self.polygon.points.insert(i, p)
        self.polygon.npoints += 1
        self.polygon.recalculate_bounds()
        if self.edgeStyles is not None:
            self.edgeStyles.insert(i - 1, self.edgeStyles[i - 1])

    def isCounterClockwise(self):
        x1 = long(self.polygon.points[-1].x)
        y1 = long(self.polygon.points[-1].y)
        area = 0
        for p in self.polygon.points:
            x2 = long(p.x)
            y2 = long(p.y)
            area += x1 * y2 - x2 * y1
            x1, y1 = x2, y2
        return area > 0

    def checkWrapping(self, mapsize):
        xwrap = self.polygon.points[0].x
        xwrap -= xwrap % mapsize.width
        ywrap = self.polygon.points[0].y
        ywrap -= ywrap % mapsize.height
        if xwrap != 0 or ywrap != 0:
            for p in self.polygon.points:
                p.x -= xwrap
                p.y -= ywrap
            self.polygon.recalculate_bounds()

    def printXml(self, file):
        if self.polygon.npoints < 3:
            raise "Invalid polygon with less than 3 points"

        if self.type == self.TYPE_BALLAREA:
            file.write('<BallArea>\n')
        elif self.type == self.TYPE_BALLTARGET:
            file.write('<BallTarget team="%d">\n' % self.team)
        elif self.type == self.TYPE_DECORATION:
            file.write('<Decor>\n')

        points = self.polygon.points[:]
        es = self.edgeStyles
        cls = None
        if es is None:
            es = [None] * len(points)
        file.write('<Polygon x="%d" y="%d" style="%s">\n' % (
                   points[0].x, points[0].y, self.style.id))

        if not self.isCounterClockwise():
            points.reverse()
            es = es[:]
            es.reverse()
        else:
            points.append(points.pop(0))

        pp = points[-1]
        for p, e in zip(points, es):
            file.write('<Offset x="%d" y="%d"' % (p.x - pp.x, p.y - pp.y))
            if e is not cls:
                if e is None:
                    id = self.style.defaultEdgeStyle.id
                else:
                    id = e.id
                file.write(' style="%s"' % e)
            file.write('/>\n')
            pp = p
        file.write('</Polygon>\n')

        if self.type == self.TYPE_BALLAREA:
            file.write('</BallArea>\n')
        elif self.type == self.TYPE_BALLTARGET:
            file.write('</BallTarget>\n')
        elif self.type == self.TYPE_DECORATION:
            file.write('</Decor>\n')


# !@# Can create starting points with incorrect wrapping
class CreateHandler(MouseEventHandler):
    def __init__(self, object):
        self.object = object
        self.poly = None
        self.remove = 0

    def mousePressed(self, me):
        c = me.canvas
        gc = c.previewgc
        if self.poly is None:
            self.poly = Polygon()
            first = me.point
            self.poly.addPoint(first.x, first.y)
            self.poly.addPoint(first.x, first.y)
            return
        c = me.canvas
        if self.remove:
            c.drawShape(self.poly, gc)
            self.remove = 0
        latest = me.point
        self.poly.points[self.poly.npoints - 1] = latest
        if me.button == 3:
            if len(self.poly.points) > 2:
                self.poly.recalculate_bounds()
                self.object.polygon = self.poly
                self.object.style = c.getModel().getDefaultPolygonStyle()
                self.object.checkWrapping(c.model.options.size)
                c.getModel().addToFront(self.object)
            c.setCanvasEventHandler(None)
            c.repaint()
            c.saveUndo()
            return
        self.poly.addPoint(latest.x, latest.y)
        c.drawShape(self.poly, gc)
        self.remove = 1

    def mouseMoved(self, me):
        c = me.canvas
        if self.poly is None:
            return
        if self.remove:
            c.drawShape(self.poly)
            self.remove = 0
        self.poly.points[self.poly.npoints - 1] = me.point
        # Bounds not recalculated here so might not be 100% correct
        c.drawShape(self.poly)
        self.remove = 1

class PolygonPointMoveHandler(MouseEventHandler):
    def __init__(self, obj, me, index, wrap):
        self.obj = obj
        self.index = index
        self.virgin = 1
        self.wrap = wrap
        ip = index - 1 % obj.polygon.npoints
        prev = obj.polygon.points[(index - 1) % obj.polygon.npoints]
        self.l1 = Line(prev, prev)
        next = obj.polygon.points[(index + 1) % obj.polygon.npoints]
        self.l2 = Line(next, next)

    def mouseMoved(self, evt):
        c = evt.canvas
        if not self.virgin:
            self.drawPreview(c)
        else:
            self.virgin = 0
        p = evt.point
        p.x += self.wrap[0]
        p.y += self.wrap[1]
        self.l1.p2 = self.l2.p2 = evt.point
        # Bounds not recalculated...
        self.drawPreview(c)

    def mouseReleased(self, evt):
        c = evt.canvas
        p = evt.point
        p.x += self.wrap[0]
        p.y += self.wrap[1]
        self.obj.polygon.points[self.index] = evt.point
        self.obj.polygon.recalculate_bounds()
        if self.index == 0:
            self.obj.checkWrapping(c.model.options.size)
        c.setCanvasEventHandler(None)
        c.repaint()
        c.saveUndo()

    def drawPreview(self, c):
        c.drawShape(self.l1)
        c.drawShape(self.l2)



import wx
import config

class Process(wx.Process):
	def __init__(self, win, argv):
		wx.Process.__init__(self, win)
		self.Redirect()
		self.win = win
		self.cmd = ' '.join(argv)
		self.pid = None
	def run(self):
		print self.cmd
		self.pid = wx.Execute(self.cmd, wx.EXEC_ASYNC, self)
		if not self.pid:
			wx.MessageDialog(self.win,
							 "Failed to execute command \"%s\". " 
							 % self.cmd,
							 "Error", 
							 wx.OK|wx.ICON_ERROR)
	def kill(self):
		if self.pid:
			wx.Process.Kill(self.pid, wx.SIGINT)
	def OnTerminate(self, pid, status):
		if status:
			wx.MessageDialog(self.win, 
							 "Command \"%s\" exited with error code %d. "
							 % (self.cmd, status),
							 "Error",
							 wx.OK|wx.ICON_ERROR).ShowModal()

class Client(Process):
	def __init__(self, win, prog):
		Process.__init__(self, win, ())
		self.prog = prog
	def join(self, host, port, nick=None):
                if config.is_muted:
                        sound = "no"
                else:
                        sound = "yes"
		if nick:
			self.cmd = "%s -name \"%s\" -port %d -join %s -sound %s" % (self.prog, 
								  nick, port,
                                                                  host, sound)
		else:
			self.cmd = "%s -port %d -join %s -sound %s" % (self.prog, port,
                                        host, sound)
		self.run()




from javastuff import *

class PolygonStyle:
    def __init__(self):
        self.visible = 1
        self.visibleInRadar = 1
        self.color = 0xFFFFFF
        self.texture = None
        self.filled = 0
        self.textured = 0
        self.defaultEdgeStyle = None
        id = ""

    def setFromFlags(self, flags):
        self.filled = not not flags & 1
        self.textured = not not flags & 2
        self.visible = not flags & 4
        self.visibleInRadar = not flags & 8

    def computeFlags(self):
        flags = 0
        if self.filled:
            flags |= 1
        if self.textured:
            flags |= 2
        if not self.visible:
            flags |= 4
        if not self.visibleInRadar:
            flags |= 8
        return flags

    def printXml(self, file):
        file.write('<Polystyle id="%s"' % self.id)
        file.write(' color="%s"' % toRGB(self.color))
        if self.texture != None:
            file.write(' texture="%s"' % self.texture.fileName)
        file.write(' defedge="%s" flags="%s"/>\n' % (
            self.defaultEdgeStyle.id, self.computeFlags()))



class Pixmap:
    def __init__(self, bmstore, filename):
        self.scalable = 1
        self.fileName = filename
        self.image = bmstore[filename]

    def printXml(self, file):
        file.write('<BmpStyle id="%s" filename="%s" flags="%d"/>\n' %
                   (self.fileName, self.fileName, self.scalable > 0))



from MapObject import MapObject

class MapBase(MapObject):
    LEFT = 0
    DOWN = 1
    RIGHT = 2
    UP = 3

    def __init__(self, x = 0, y = 0, dir = 32, team = 1, bmstore = None):
        if dir < 16:
            imgname = 'base_left.gif'
        elif dir < 48:
            imgname = 'base_down.gif'
        elif dir < 80:
            imgname = 'base_right.gif'
        elif dir < 112:
            imgname = 'base_up.gif'
        else:
            imgname = 'base_left.gif'
        MapObject.__init__(self, x, y, 35 * 64, 35 * 64, bmstore[imgname])
        self.dir = dir
        self.team = team

    def printXml(self, file):
        file.write('<Base x="%d" y="%d" team="%d" dir="%d"/>\n' % (
            self.x, self.y, self.team, self.dir))



from __future__ import nested_scopes
from LineStyle import LineStyle
from PolygonStyle import PolygonStyle
from MapOptions import MapOptions
from MapBall import MapBall
from MapPolygon import MapPolygon
from MapBase import MapBase
from MapFuel import MapFuel
from MapCheckPoint import MapCheckPoint
from Pixmap import Pixmap
from javastuff import *
import xml.sax

class Poly:
    def __init__(self):
        self. hasSpecialEdges = 0
        self.points = []
        self.type = MapPolygon.TYPE_NORMAL
        self.team = 1

class PolyPoint:
    def __init__(self, x, y, style):
        self.x = int(x)
        self.y = int(y)
        self.style = style

class PolyStyle:
    def __init__(self, id, color, textureId, defEdgeId, flags):
        self.id = id
        self.color = color
        self.textureId = textureId
        self.defEdgeId = defEdgeId
        self.flags = flags

class MapModel:
    def __init__(self):
        self.defaultScale = 1./64
        self.objects = []
        self.pixmaps = []
        self.polyStyles = []
        self.edgeStyles = []
        self.options = MapOptions()

        ls = LineStyle("internal", 0, 0xFFFF, LineStyle.STYLE_INTERNAL)
        self.edgeStyles.append(ls)
        ls = LineStyle("default", 0, 0xFF, LineStyle.STYLE_SOLID)
        self.edgeStyles.append(ls)
        self.defEdgeStyleIndex = 1
        ps = PolygonStyle()
        ps.id = "default"
        ps.visible = ps.visibleInRadar= 1
        ps.filled = ps.textured = 0
        ps.defaultEdgeStyle = ls
        self.polyStyles.append(ps)
        self.defPolygonStyleIndex = 1

    def getDefaultScale(self):
        return self.defaultScale

    def setDefaultScale(self, ds):
        self.defaultScale = ds

    def getDefaultPolygonStyle(self):
        if self.defPolygonStyleIndex >= len(self.polyStyles):
            self.defPolygonStyleIndex = len(self.polyStyles) - 1
        return self.polyStyles[self.defPolygonStyleIndex]

    def setDefaultPolygonStyle(self, index):
        self.defPolygonStyleIndex = index

    def load(self, file, bmstore):
        self.edgeStyles = []
        ls = LineStyle("internal", 0, 0xFFFF, LineStyle.STYLE_INTERNAL)
        self.edgeStyles.append(ls)
        self.polyStyles = []

        xml.sax.parse(file, MapDocumentHandler(self, bmstore))
        self.defEdgeStyleIndex = len(self.edgeStyles) - 1
        self.defPolygonStyleIndex = len(self.polyStyles) - 1

    def save(self, file):
        file.write('<XPilotMap>\n')
        self.options.printXml(file)
        for pm in self.pixmaps:
            pm.printXml(file)
        for es in self.edgeStyles:
            es.printXml(file)
        for ps in self.polyStyles:
            ps.printXml(file)
        o = self.objects[:]
        o.reverse()
        for obj in o:
            obj.printXml(file)
        file.write('</XPilotMap>\n')
        file.close()

    def addToFront(self, moNew):
        self.addObject(moNew, 1)

    def addToBack(self, moNew):
        self.addObject(moNew, 0)

    def addObject(self, moNew, front):
        znew = moNew.getZOrder()
        for i in range(len(self.objects)):
            zold = self.objects[i].getZOrder()
            if znew < zold or front and znew == zold:
                self.objects.insert(i, moNew)
                return
        self.objects.append(moNew)

    def removeObject(self, mo):
        self.objects.remove(mo)

class MapParseError(Exception):
    pass

class MapDocumentHandler(xml.sax.ContentHandler):
    def __init__(self, model, bmstore):
        self.model = model
        self.bmstore = bmstore
        self.polys = []
        self.estyles = {}
        self.estyles["internal"] = model.edgeStyles[0]
        self.pstyles = {}
        self.bstyles = {}
        self.opMap = {}
        self.ballArea = 0
        self.ballTargetTeam = -1
        self.decor = 0

    def startElement(self, name, atts):
        bm = self.bmstore
        name = name.lower()
        def getxy():
            return int(atts["x"]), int(atts["y"])
        def getint(name, base = 10):
            return int(atts[name], base)
        def addobj(o):
            self.model.addToFront(o)
        if name == "polygon":
            poly = Poly()
            poly.style = atts["style"]
            poly.points.append(PolyPoint(atts["x"], atts["y"], None))
            if self.ballArea:
                poly.type = MapPolygon.TYPE_BALLAREA
            elif self.ballTargetTeam != -1:
                poly.type = MapPolygon.TYPE_BALLTARGET
                poly.team = self.ballTargetTeam
            elif self.decor:
                poly.type = MapPolygon.TYPE_DECORATION
            self.poly = poly
        elif name == "offset":
            es = atts.get("style")
            if es != None:
                self.poly.hasSpecialEdges = 1
            self.poly.points.append(PolyPoint(atts["x"], atts["y"], es))
        elif name == 'edgestyle':
            id = atts["id"]
            color = getint("color", 16)
            width = getint("width")
            style = getint("style")
            if width == -1:
                width = 0
                style = LineStyle.STYLE_HIDDEN
            ls = LineStyle(id, width, color, style)
            self.estyles[id] = ls
            self.model.edgeStyles.append(ls)
        elif name == 'polystyle':
            id = atts["id"]
            cstr = atts.get("color")
            if cstr is not None:
                col = int(cstr, 16)
            else:
                col = None
            flagstr = atts.get('flags', '1')
            flags = int(flagstr)
            self.pstyles[id] = PolyStyle(id, col, atts.get("texture"),
                                         atts["defedge"], flags)
        elif name == 'bmpstyle':
            id = atts["id"]
            fileName = atts["filename"]
            flags = getint("flags")
            pm = Pixmap(self.bmstore, fileName)
            pm.fileName = fileName
            pm.scalable = flags != 0
            self.bstyles[id] = pm
            self.model.pixmaps.append(pm)
        elif name == 'option':
            self.opMap[atts["name"]] = atts["value"]
        elif name == 'fuel':
            x, y = getxy()
            addobj(MapFuel(x, y, bm))
        elif name == 'ball':
            x, y = getxy()
            team = getint("team")
            addobj(MapBall(x, y, team, bm))
        elif name == 'base':
            x, y = getxy()
            dir = getint("dir")
            team = getint("team")
            addobj(MapBase(x, y, dir, team, bm))
        elif name == 'check':
            x,y = getxy()
            addobj(MapCheckPoint(x, y, bm))
        elif name == 'ballarea':
            self.ballArea = 1
        elif name == 'balltarget':
            self.ballTargetTeam = getint("team")
        elif name == 'decor':
            decor = 1
        elif name == 'xpilotmap':
            pass
        elif name == 'generaloptions':
            pass
        else:
            print "Tag: ", name
            raise MapParseError("Unknown tag")

    def endElement(self, name):
        name = name.lower()
        if name == 'polygon':
            self.polys.append(self.poly)
        elif name == 'ballarea':
            self.ballArea = 0
        elif name == 'balltarget':
            self.ballTargetTeam = -1
        elif name == 'decor':
            self.decor = 0

    def endDocument(self):
        self.model.options = MapOptions(self.opMap)
        for ps in self.pstyles.values():
            style = PolygonStyle()
            style.id = ps.id
            style.setFromFlags(ps.flags)
            if style.textured:
                if ps.textureId is None:
                    raise MapParseError("Textured polygon, no texture name")
                style.texture = self.bstyles[ps.textureId]
            if ps.color is None:
                style.color = 0xFFFFFF
            else:
                style.color = ps.color
            ls = self.estyles[ps.defEdgeId]
            style.defaultEdgeStyle = ls
            self.model.polyStyles.append(style)
            ps.ref = style
        for p in self.polys:
            ps = self.pstyles[p.style]
            style = ps.ref
            defls = style.defaultEdgeStyle
            if p.hasSpecialEdges:
                edges = []
            else:
                edges = None
            awtp = Polygon()
            pnt = p.points[0]
            x = pnt.x
            y = pnt.y
            awtp.addPoint(x, y)
            for pnt in p.points[1:-1]:
                x += pnt.x
                y += pnt.y
                awtp.addPoint(x, y)
            cls = None
            if edges != None:
                for pnt in p.points[1:]:
                    if pnt.style != None:
                        cls = self.estyles[pnt.style]
                        if cls == defls:
                            cls = None
                    edges.append(cls)
            mp = MapPolygon(awtp, style, edges)
            mp.type = p.type
            mp.team = p.team
            self.model.addToFront(mp)




# Editor will search for images (both the ones it uses by default such
# as checkpoint bitmaps and those named in map files) from these
# paths. One path should probably point to ../images/ (as stored in
# CVS) from this directory, to find the basic images (checkpoint.gif
# etc). Other paths are needed only if you edit maps with custom textures.

PIXMAP_PATH = ["/home/uau/reiser/xpilot/contrib/jxpmap/images/",
               "/home/uau/games/lib/xpilot/textures/rtc/"]

# Currently undo is done in a simple way, storing a copy of the whole
# map each time certain operations are done (supposedly those you
# might wish to undo). This takes memory and due to the inefficient
# copying algorithm can also take a visible amount of time (depending
# on map complexity and computer speed).
# This limits the number of copies done. Setting it lower means less memory
# used (but doesn't affect the time taken); setting to 0 means undo is
# disabled (and thus no speed penalty either).

UNDO_LEVELS = 100



import os
import sys

def which(app):
	# Does this work on Windows?
	dirnames = os.environ['PATH'].split(os.path.pathsep)
	for dirname in dirnames:
		filename = os.path.join(dirname, app)
		if (os.access(filename, os.X_OK) and
		    os.path.isfile(filename)):
			return filename
	return os.environ.get(app.replace('-', '_'), None)

meta = ('meta.xpilot.org', 4401)
client_x11 = which('xpilot-ng-x11')
client_sdl = which('xpilot-ng-sdl')
if client_sdl:
	client = client_sdl
else:
	client = client_x11
if (sys.platform == 'win32'):
	server = r'"C:\Program Files\XPilotNG-SDL\xpilot-ng-server"'
else:
	server = which('xpilot-ng-server')
if (sys.platform == 'win32'):
	xpilotrc = r'C:\Progra~1\XPilotNG-SDL\xpilotrc.txt'
else:
	xpilotrc = os.path.expanduser('~/.xpilotrc')
record_url = 'http://xpilot.sourceforge.net/ballruns/'
xpreplay = which('xpilot-ng-replay')
jxpmap_url = 'http://xpilot.sf.net/jxpmap/jxpmap.jnlp'
javaws = which('javaws')
mapedit = which('xpilot-ng-xp-mapedit')
irc_server = 'irc.freenode.net'
irc_channel = '#xpilot'
if (sys.platform == 'win32'):
	mapdir = r'C:\Progra~1\XPilotNG-SDL\lib\maps'
else:
	mapdir = '/usr/local/share/xpilot-ng/maps'
png_path = os.path.dirname(os.path.abspath(__file__))
is_muted = False



from MapModel import MapModel
from MapCanvas import MapCanvas
from MapFuel import MapFuel
from MapPolygon import MapPolygon
from MapBase import MapBase
from MapBall import MapBall
from MapCheckPoint import MapCheckPoint
import gtk
import GDK

class MainFrame:
    def __init__(self):
        self.canvas = None
        self.window = gtk.GtkWindow(gtk.WINDOW_TOPLEVEL)
        self.window.connect("delete_event", self.delete_event)
        self.window.set_title("jXPMap Editor")
        self.box0 = gtk.GtkVBox(gtk.FALSE, 0)
        menubar = self.buildMenuBar()
        self.box0.pack_start(menubar, gtk.FALSE, gtk.FALSE, 0)
        self.box1 = gtk.GtkHBox(gtk.FALSE, 0)
        self.box0.pack_start(self.box1, gtk.TRUE, gtk.TRUE, 0)
        self.window.add(self.box0)
        self.buildToolBar()
        #self.ActionMap()
        #self.buildInputMap()

        # The visual calls are needed to make sure drawing bitmaps
        # (draw_*_image) works
        # ... at least according to some documentation that doesn't look
        # entirely reliable

        gtk.push_rgb_visual()
        area = gtk.GtkDrawingArea()
        gtk.pop_visual()
        area.size(500, 500)
        self.box1.pack_start(area, gtk.TRUE, gtk.TRUE, 0)
        area.connect("expose-event", self.area_expose_cb)

        # Change the default background to black so that the window
        # doesn't flash badly when resizing

        style = area.get_style().copy()
        style.bg[gtk.STATE_NORMAL] = area.get_colormap().alloc(0, 0, 0)
        area.set_style(style)
        area.connect("button_press_event", self.mouse_event)
        area.connect("button_release_event", self.mouse_event)
        area.connect("motion_notify_event", self.mouse_event)
        area.connect("configure_event", self.configure_event)
        area.set_events(GDK.EXPOSURE_MASK | GDK.BUTTON_PRESS_MASK |
                        GDK.BUTTON_RELEASE_MASK | GDK.POINTER_MOTION_MASK |
                        GDK.POINTER_MOTION_HINT_MASK)

        # The HINT_MASK means there won't be a callback for every mouse
        # motion event, only after certain other events or after an explicit
        # reading of the location with x, y = window.pointer.
        # This helps to avoid getting a queue of events if they're not handled
        # quickly enough.

        area.show()
        self.zoom = 0
        self.box1.show()
        self.box0.show()
        self.window.show()
        self.canvas = MapCanvas(area)

    def mouse_event(self, area, event):
        if event.type == GDK.BUTTON_PRESS and event.button == 5:
            self.canvas.setCenter(event.x, event.y)
            self.canvas.paint(None)
        else:
            self.canvas.handle(area, event)
        return gtk.TRUE

    def area_expose_cb(self, area, event):
        self.canvas.paint(event.area)
        return gtk.TRUE

    def delete_event(self, widget, event):
        gtk.mainquit()
        return gtk.FALSE

    def configure_event(self, widget, event):
        if self.canvas is not None:
            self.canvas.resize()

    def buildMenuBar(self):
        menu_items = (
            ("/File", None, None, 0, "<Branch>"),
            ("/File/New", None, self.newMap, 0, None),
            ("/File/Open", None, self.openMap, 0, None),
            ("/File/Save", None, self.saveMap, 0, None),
            )
        accel_group = gtk.GtkAccelGroup()
        item_factory = gtk.GtkItemFactory(gtk.GtkMenuBar.get_type(), "<main>",
                                          accel_group)
        item_factory.create_items(menu_items)
        menubar = item_factory.get_widget("<main>")
        menubar.show()
        return menubar

    def buildToolBar(self):
        box = gtk.GtkVBox(gtk.FALSE, 0)
        l = [["New polygon", MapPolygon], ["New fuel", MapFuel],
             ["New checkpoint", MapCheckPoint], ["New ball", MapBall],
             ["New base", MapBase]]
        l = [[a, self.newObject, b] for a, b in l]
        l.extend([["Erase", self.toggleEraseMode, None],
                  ["Toggle points", self.canvasToggle, 'showPoints'],
                  ["Toggle filled", self.canvasToggle, 'filled'],
                  ["Toggle textured", self.canvasToggle, 'textured'],
                  ["Undo", self.undo, -1], ["Redo", self.undo, 1],
                  ["Zoom in", self.zoom, 1], ["Zoom out", self.zoom, -1]
                  ])

        for name, callback, data in l:
            button = gtk.GtkButton(name)
            button.connect("clicked", callback, data)
            box.pack_start(button, gtk.FALSE, gtk.FALSE, 0)
            button.show()
        self.zoomlabel = gtk.GtkLabel("1.00")
        self.zoomlabel.show()
        box.pack_start(self.zoomlabel, gtk.FALSE, gtk.FALSE, 0)
        self.box1.pack_start(box, gtk.FALSE, gtk.FALSE, 0)
        box.show()

    def undo(self, widget, value):
        self.canvas.undo(value)

    def newObject(self, widget, Type):
        self.canvas.setErase(0)
        self.canvas.setCanvasEventHandler(Type(bmstore = self.canvas.bmstore).
                                          getCreateHandler())

    def zoom(self, widget, value):
        self.zoom += value
        if self.zoom > 20:
            self.zoom = 20
            return
        elif self.zoom < -15:
            self.zoom = -15
            return
        self.updateScale()

    def canvasToggle(self, widget, data):
        setattr(self.canvas, data, not getattr(self.canvas, data))
        self.canvas.repaint()

    def toggleEraseMode(self, widget, data):
        self.canvas.setCanvasEventHandler(None)
        self.canvas.setErase(not self.canvas.isErase())

    def updateScale(self):
        z = 1.3**self.zoom
        self.zoomlabel.set_text("%.2f" % z)
        z *= self.canvas.getModel().getDefaultScale()
        self.canvas.setScale(z)
        self.canvas.repaint()

    def setModel(self, model):
        self.canvas.setModel(model)

    def newMap(self, menuitem = None, data = None):
        self.mapFile = None
        self.setModel(MapModel())
        #showOptions()

    def openMap(self, menuitem, data):
        print 'Not opening map'

    def saveMap(self, menuitem, data):
        file = open('tempfile', 'w')
        self.canvas.getModel().save(file)

    def main(self):
        gtk.mainloop()

if __name__ == '__main__':
    mf = MainFrame()
    import sys
    if len(sys.argv) <= 1:
        mf.newMap()
    else:
        model = MapModel()
        file = open(sys.argv[1])
        model.load(file, mf.canvas.bmstore)
        mf.setModel(model)
    mf.main()



import wx
import irclib
import threading
import wx.lib.newevent
import time

(AppendMsgEvent, EVT_APPEND_MSG) = wx.lib.newevent.NewEvent()

class IrcPanel(wx.Panel):
	def __init__(self, parent, server, nick, channel):
		wx.Panel.__init__(self, parent, -1)
		p = wx.Panel(self)
		ps = wx.BoxSizer(wx.HORIZONTAL)
		p.SetSizer(ps)
		l = wx.StaticText(p, -1, "Message:")
		l.SetForegroundColour(wx.Color(255,255,255))
		ps.Add(l, 0, wx.ALIGN_CENTER_VERTICAL|wx.ALL, 5)
		self.message = wx.TextCtrl(p, -1, '', style=wx.TE_PROCESS_ENTER)
		self.Bind(wx.EVT_TEXT_ENTER, self.on_say, self.message)
		ps.Add(self.message, wx.EXPAND, wx.ALIGN_CENTER_VERTICAL|wx.ALL, 5)
		b = wx.Button(p, -1, "Say")
		self.Bind(wx.EVT_BUTTON, self.on_say, b)
		ps.Add(b, 0, wx.ALL, 5)
		sz = wx.BoxSizer(wx.VERTICAL)
		self.SetSizer(sz)
		self.area = wx.TextCtrl(self, -1, style=wx.TE_READONLY|wx.TE_MULTILINE)
		sz.Add(self.area, wx.EXPAND, wx.EXPAND|wx.ALL, 5)
		sz.Add(p, 0, wx.EXPAND|wx.ALL, 5)
		self.Bind(EVT_APPEND_MSG, self.on_append)
		self.worker = WorkerThread(self, server, nick, channel)
		self.worker.start()
	def on_say(self, evt):
		self.worker.send_message(self.message.GetValue())
		self.message.SetValue('')
	def on_append(self, evt):
		self.area.AppendText("[%s] %s\n" 
							 % (time.strftime("%H:%M:%S"), evt.message))
	def append(self, str):
		wx.PostEvent(self, AppendMsgEvent(message = str))
	def Destroy(self):
		self.worker.kill()
		wx.Panel.Destroy(self)

class WorkerThread(threading.Thread):
	def __init__(self, panel, server, nick, channel):
		threading.Thread.__init__(self)
		self.ui = panel
		self.server = server
		self.nick = nick
		self.realnick = nick
		self.channel = channel
		self.index = 0
		self.namreply = ''
		self.die = False
	def kill(self):
		self.die = True
		self.connection.quit()
	def run(self):
		irc = irclib.IRC()
		for m in filter(lambda x: x.startswith('on_'), dir(self)):
			irc.add_global_handler(m[3:], getattr(self, m))
		c = irc.server()
		self.connection = c
		c.connect(self.server, 6667, self.nick)
		while not self.die:
			irc.process_once(0.2)
		c.disconnect()
	def send_message(self, str):
		self.connection.privmsg(self.channel, str)
		msg = "%s: %s" % (self.realnick, str)
		self.ui.append(msg)
	def on_welcome(self, c, evt):
		self.ui.append("Welcome to channel %s on %s. Here you can ask questions about xpilot or just chat with the friendly xpilot people." % (self.channel, self.server))
		c.join(self.channel)
	def on_nicknameinuse(self, c, evt):
		self.index += 1
		self.realnick = "%s_%d" % (self.nick, self.index)
		c.nick(self.realnick)
	def on_namreply(self, c, evt):
		self.namreply += evt.arguments()[2]
	def on_endofnames(self, c, evt):
		msg = "Currently online: %s" % self.namreply
		self.namreply = ''
		self.ui.append(msg)
	def on_privmsg(self, c, evt):
		msg = "%s->%s: %s" % (irclib.nm_to_n(evt.source()), 
							  evt.target(), evt.arguments()[0])
		self.ui.append(msg)
	def on_pubmsg(self, c, evt):
		msg = "%s: %s" % (irclib.nm_to_n(evt.source()), evt.arguments()[0])
		self.ui.append(msg)
	def on_join(self, c, evt):
		msg = "%s has joined %s" % (irclib.nm_to_n(evt.source()), evt.target())
		self.ui.append(msg)
	def on_part(self, c, evt):
		msg = "%s has left %s" % (irclib.nm_to_n(evt.source()), evt.target())
		self.ui.append(msg)
	def on_quit(self, c, evt):
		msg = "%s has quit" % irclib.nm_to_n(evt.source())
		self.ui.append(msg)

class Frame(wx.Frame):
    def __init__(
		self, parent, ID, title, pos=wx.DefaultPosition,
		size=(800,600), style=wx.DEFAULT_FRAME_STYLE):
		wx.Frame.__init__(self, parent, ID, title, pos, size, style)
		box = wx.BoxSizer(wx.VERTICAL)
		box.Add(IrcPanel(self, "irc.freenode.net", "foobar", "#xpilottest"),
				wx.EXPAND, wx.EXPAND, 0, 0)
		self.SetSizer(box)

class App(wx.App):
	def OnInit(self):
		frame = Frame(None, -1, "")
		self.SetTopWindow(frame)
		frame.Show(True)
		return True

if __name__ == '__main__':
	app = App(0)
	app.MainLoop()



#!/usr/bin/env python

import wxversion
try:
	wxversion.select('2.6')
except wxversion.VersionError:
	try:
		wxversion.select('2.4')
	except wxversion.VersionError:
		wxversion.select('2.5')
import wx
import wx.html as html
import os
import urllib
import metaui
import config
import options
import ircui
import xputil
import serverui

nick = None
def get_nick():
	global nick
	if nick: return nick
	if os.path.exists(config.xpilotrc):
		opts = options.parse_xpilotrc(config.xpilotrc)
		try:
			nick = opts['name']
		except KeyError:
			pass
	if nick: return nick
	try:
		nick = os.environ['USER']
	except KeyError:
		nick = os.environ.get('USERNAME', 'xpilot-user')
	return nick

class RecordingsPanel(html.HtmlWindow):
	def __init__(self, parent):
		html.HtmlWindow.__init__(self, parent)
		self.baseurl = config.record_url
		self.LoadPage(self.baseurl)
	def OnLinkClicked(self, link):
		file = urllib.urlretrieve(self.baseurl + link.GetHref())[0]
		xputil.Process(self, (config.xpreplay, file)).run()

class MenuPanel(wx.Panel):
	def __init__(self, parent, items):
		wx.Panel.__init__(self, parent)
		self.SetBackgroundColour(wx.BLACK)
		self.frame = parent
		p = wx.Panel(self)
		p.SetBackgroundColour(wx.BLACK)
		s = wx.BoxSizer(wx.VERTICAL)
		p.SetSizer(s)
		items.insert(0, ())
		for item in items:
			if len(item):
				b = wx.Button(p, -1, item[0])
				b.SetSize((300, 50))
				b.SetForegroundColour(wx.Colour(255, 255, 255))
				b.SetBackgroundColour(wx.Colour(50, 50, 204))
				b.SetFont(wx.Font(20, wx.DEFAULT, wx.SLANT, wx.NORMAL, 0, ""))
				if item[1]:
					self.Bind(wx.EVT_BUTTON, item[1], b)
				else:
					b.Disable()
				s.Add(b, 0, wx.ALL|wx.EXPAND|wx.ALIGN_CENTER_HORIZONTAL, 10)
			else:
				s.Add((0,30), 0, wx.ALL, 10)
		s = wx.BoxSizer(wx.VERTICAL)
		self.SetSizer(s)
		s.Add(p, wx.EXPAND, wx.ALIGN_CENTER_HORIZONTAL, 0)
		s.Add(self.makeBottomPanel(), 0, wx.EXPAND, 0)	
	def makeBottomPanel(self):
		p = wx.Panel(self)
		p.SetBackgroundColour(wx.BLACK)
		s = wx.FlexGridSizer(1, 3)
		s.Add(wx.StaticBitmap(
			p, -1,
			wx.Bitmap(os.path.join(config.png_path, "swgrid.png"),
			wx.BITMAP_TYPE_ANY)), 0, 0, 0)
		s.Add((0,0), 0, 0, 0)
		s.Add(wx.StaticBitmap(
			p, -1,
			wx.Bitmap(os.path.join(config.png_path, "segrid.png"),
			wx.BITMAP_TYPE_ANY)), 0, 0, 0)
		s.AddGrowableCol(1);
		p.SetSizer(s)
		return p
	def show(self, panel):
		self.frame.setContentPanel(panel)

class  MapEditorMenu(MenuPanel):
	def __init__(self, parent):
		b = []
		if config.javaws:
			b.append(("  Polygon map editor  ", self.onPoly))
		if config.mapedit:
			b.append(("Block map editor", self.onBlock))
		if b:
			MenuPanel.__init__(self, parent, b)
		else:
			# FIXME: raise some error; we shouldn't even create this menu if
			# we have nothing to put in it
			pass
	def onPoly(self, evt):
		xputil.Process(self, (config.javaws, config.jxpmap_url)).run()
	def onBlock(self, evt):
		xputil.Process(self, (config.mapedit,)).run()

class ToolsMenu(MenuPanel):
	def __init__(self, parent):
		b = []
		if config.client:
			b.append(("  Client configuration  ", self.onClientConfig))
		if config.xpreplay:
			b.append(("XP-Replay", self.onXPReplay))
			b.append(("Recordings", self.onRecordings))
		if config.mapedit and config.javaws:
			b.append(("Map editor", self.onMapEditor))
		elif config.mapedit:
			b.append(("Block map editor", self.onBlock))
		elif config.javaws:
			b.append(("Polygon map editor", self.onPoly))
		MenuPanel.__init__(self, parent, b)
	def onMapEditor(self, evt):
		self.show(MapEditorMenu(self.frame))
	def onRecordings(self, evt):
		self.show(RecordingsPanel(self.frame))
	def onXPReplay(self, evt):
		dlg = wx.FileDialog(
            None, message="Choose a recording", defaultDir=os.getcwd(), 
            defaultFile="", wildcard="*.xpr*", style=wx.OPEN | wx.CHANGE_DIR)
		if dlg.ShowModal() == wx.ID_OK:
			xputil.Process(self, (config.xpreplay, dlg.GetPath())).run()
	def onClientConfig(self, evt):
		self.show(options.ClientOptionsPanel(
				self.frame, config.client, config.xpilotrc))
	def onPoly(self, evt):
		xputil.Process(self, (config.javaws, config.jxpmap_url)).run()
	def onBlock(self, evt):
		xputil.Process(self, (config.mapedit,)).run()

class MainMenu(MenuPanel):
	def __init__(self, parent):
		b = []
		b.append(("    Internet servers    ", self.onInternet))
		if config.server:
			b.append(("Start server", self.onStart))
		if config.client or config.xpreplay or config.mapedit or config.javaws:
			b.append(("Tools", self.onTools))
		b.append(("Support and Chat", self.onChat))
# FIXME: This should be a fullscreen widget in the corner instead.
#		b.append(("Windowed", self.onWindowed))
		b.append(("Quit", self.onQuit))
		MenuPanel.__init__(self, parent, b)
	def onInternet(self, evt):
		meta = metaui.Panel(self.frame,
							config.meta,
							xputil.Client(self, config.client))
		self.show(meta)
		meta.RefreshList()
	def onTools(self, evt):
		self.show(ToolsMenu(self.frame))
	def onQuit(self, evt):
		self.frame.Close()
	def onStart(self, evt):
		self.show(options.ServerOptionsPanel(
				self.frame, 
				xputil.Client(self, config.client), 
				config.server,
				config.mapdir))
	def onChat(self, evt):
		self.show(ircui.IrcPanel(self.frame, config.irc_server, get_nick(), 
								 config.irc_channel))
	def onWindowed(self, evt):
		# FIXME: assert self.frame.fullscreen=True
		# FIXME: remove "Windowed" button & replace with "Fullscreen"
		self.frame.ShowFullScreen(False)
		self.frame.fullscreen=False
	def onFullscreen(self, evt):
		# FIXME: assert self.frame.fullscreen=False
		# FIXME: remove "Fullscreen" button & replace with "Windowed"
		self.frame.ShowFullScreen(True)
		self.frame.fullscreen=True

class MainFrame(wx.Frame):
	def __init__(self, *args, **kwds):
		kwds["style"] = wx.DEFAULT_FRAME_STYLE
		wx.Frame.__init__(self, *args, **kwds)
		self.SetSize((800, 650))
		self.SetBackgroundColour(wx.BLACK)
		s = wx.BoxSizer(wx.VERTICAL)
		s.Add(self.makeTopPanel(), 0, wx.EXPAND, 0)
		self.SetAutoLayout(True)
		self.SetSizer(s)
		self.history = []
		self.contentPanel = None
		self.setContentPanel(MainMenu(self))
		self.fullscreen = None
	def setContentPanel(self, p):
		if self.contentPanel:
			self.GetSizer().Detach(self.contentPanel)
			# reusing panel instances causes flickering
			# so I save the class instead of the instance
			self.history.append(self.contentPanel.__class__)
			self.contentPanel.Destroy()
		self.contentPanel = p
		self.GetSizer().Insert(1, p, wx.EXPAND, wx.EXPAND)
		self.backButton.Show(len(self.history) > 0)
		self.Layout()
	def makeTopPanel(self):
		p = wx.Panel(self)
		p.SetBackgroundColour(wx.BLACK)
		s = wx.FlexGridSizer(1, 5)
		s.Add(wx.StaticBitmap(
			p, -1,
			wx.Bitmap(os.path.join(config.png_path, "nwgrid.png"),wx.BITMAP_TYPE_ANY)),
			0, 0, 0)
		s.Add((0,0), 0, 0, 0)
		s.Add(wx.StaticBitmap(
			p, -1,
			wx.Bitmap(os.path.join(config.png_path, "logo.png"),wx.BITMAP_TYPE_ANY)),
			0, wx.ALIGN_CENTER_VERTICAL|wx.ALL, 1)
		s.Add((0,0), 0, 0, 0)
		s.Add(wx.StaticBitmap(
			p, -1,
			wx.Bitmap(os.path.join(config.png_path, "negrid.png"), wx.BITMAP_TYPE_ANY)), 
			0, 0, 0)
		s.AddGrowableCol(1);
		s.AddGrowableCol(3);
		p.SetSizer(s)
		b = wx.Button(p, -1, "Back")
		b.SetForegroundColour(wx.Colour(255, 255, 255))
		b.SetBackgroundColour(wx.Colour(50, 50, 204))
		b.SetPosition((10,10))
		b.Bind(wx.EVT_BUTTON, self.onBack, b)
		b.Show(False)
		self.backButton = b
		return p
	def onBack(self, evt):
		if len(self.history) == 0:
			return
		old = self.contentPanel
		self.GetSizer().Detach(old)
		old.Destroy()
		pc = self.history.pop()
		p = pc(self)
		self.contentPanel = p
		self.GetSizer().Insert(1, p, wx.EXPAND, wx.EXPAND)
		self.backButton.Show(len(self.history) > 0)
		self.Layout()

class App(wx.App):
	def OnInit(self):
		frame = MainFrame(None, -1, "XPilot NG Control Center")
		self.SetTopWindow(frame)
		frame.Show(True)
# FIXME: We can only default to fullscreen when we support switching back
#        and forth between fullscreen & windowed.
#		frame.fullscreen=True
#		frame.ShowFullScreen(True)
		return True

def main():
	app = App(0)
	app.MainLoop()

if __name__ == '__main__':
	main()



from MapObject import MapObject

class MapBall(MapObject):
    def __init__(self, x = 0, y = 0, team = 1, bmstore = None):
        MapObject.__init__(self, x, y, 21 * 64, 21 * 64, bmstore["ball.gif"])
        self.team = team

    def printXml(self, file):
        file.write('<Ball x="%d" y="%d" team="%d"/>\n' % (
            self.x, self.y, self.team))



#!/usr/bin/env python
from __future__ import division, generators

KEEP_MAPDATA = 0
WALL_SHRINK_HACK = 0

# Requires python 2.2 or newer

# Some parts could be written more clearly with higher-level constructs,
# but have been optimized for speed (converting big maps can still take
# several seconds).

removedopts = """
analyticalcollisiondetection edgebounce extraborder
maxshieldedplayerwallbounceangle maxunshieldedplayerwallbounceangle
oneplayeronly scoretablefilename teamassign
""".split()

# 'mapdata' has been removed too, but isn't ignored by this script.
# 'numberofrounds', 'numrounds' and 'roundstoplay' have been used for the same
# thing

illegalopts = "noquit plockserver timerresolution".split()

knownopts = """
allowclusters allowheatseekers allowlasermodifiers allowmodifiers
allownukes allowplayerbounces allowplayercrashes allowplayerkilling
allowshields allowshipshapes allowsmartmissiles allowtorpedoes
allowviewing ballkillscoremult ballradius ballswallbounce baseminerange
cannonitemprobmult cannonsmartness cannonsuseitems checkpointradius
cloakedexhaust clusterkillscoremult contactport
crashscoremult debriswallbounce defaultsfilename denyhosts
destroyitemincollisionprob detonateitemonkillprob distinguishmissiles
dropitemonkillprob dump ecmsreprogrammines edgewrap
explosionkillscoremult firerepeatrate framespersecond friction
gameduration gravity gravityangle gravityanticlockwise
gravityclockwise gravitypoint gravitypointsource gravityvisible
heatkillscoremult help identifymines idlerun ignore20maxfps
ignore20maxfps initialafterburners initialarmor initialautopilots
initialcloaks initialdeflectors initialecms initialemergencyshields
initialemergencythrusts initialfuel initialhyperjumps initiallasers
initialmines initialmirrors initialmissiles initialphasings
initialrearshots initialsensors initialtanks initialtractorbeams
initialtransporters initialwideangles itemafterburnerprob
itemarmorprob itemautopilotprob itemcloakprob itemconcentratorprob
itemconcentratorradius itemconcentratorvisible itemdeflectorprob
itemecmprob itememergencyshieldprob itememergencythrustprob
itemenergypackprob itemhyperjumpprob itemlaserprob itemmineprob
itemmirrorprob itemmissileprob itemphasingprob itemprobmult
itemrearshotprob itemsensorprob itemswallbounce itemtankprob
itemtractorbeamprob itemtransporterprob itemwideangleprob keepshots
laserisstungun laserkillscoremult limitedlives limitedvisibility
lockotherteam loseitemdestroys mapauthor mapfilename mapheight mapname
mapwidth maxafterburners maxarmor maxautopilots maxcloaks
maxdeflectors maxecms maxemergencyshields maxemergencythrusts maxfuel
maxhyperjumps maxitemdensity maxlasers maxmines maxminesperpack
maxmirrors maxmissiles maxmissilesperpack maxobjectwallbouncespeed
maxphasings maxplayershots maxrearshots maxroundtime maxsensors
maxshieldedwallbouncespeed maxtanks maxtractorbeams maxtransporters
maxunshieldedwallbouncespeed maxvisibilitydistance maxwideangles
minefusetime minelife minescoremult minesonradar mineswallbounce
minvisibilitydistance missilelife missilesonradar missileswallbounce
movingitemprob noquit nukeclusterdamage nukeminmines nukeminsmarts
nukesonradar numberofrounds numrounds roundstoplay
objectwallbouncebrakefactor objectwallbouncelifefactor
password playerlimit playersonradar
playerstartsshielded playerwallbouncebrakefactor plockserver racelaps
recordmode reporttometaserver reset resetonhuman rogueheatprob
roguemineprob rounddelay runoverkillscoremult searchdomainforxpilot
shieldeditempickup shieldedmining shipmass shotkillscoremult shotlife
shotmass shotsgravity shotspeed shotswallbounce shovekillscoremult
smartkillscoremult sparkswallbounce tankkillscoremult targetkillteam
targetsync targetteamcollision teamcannons teamfuel teamimmunity
teamplay timerresolution timing torpedokillscoremult
treasurecollisiondestroys treasurecollisionmaykill treasurekillteam
treasuresonradar version wallbouncedestroyitemprob
wallbouncefueldrainmult worldlives wormholevisible wormtime
wreckagecollisionmaykill mapdata baselesspausing
pausedframespersecond pausedfps waitingframespersecond waitingfps
robots maxrobots minrobots robotfile robotstalk robotsleave robotleavelife
robotleavescore robotleaveratio robotteam restrictrobots reserverobotteam
robotrealname robothostname ecmsreprogramrobots
maxdefensiveitems maxoffensiveitems usewreckage
""".split()

def checkopts(options):
    dany = 0
    unknown = []
    illegal = []
    for opt in options.keys():
	if opt in removedopts:
	    if not dany:
		dany = 1
		print >> sys.stderr, "Removing the following options:"
	    print >>sys.stderr, opt + "  ",
	    del options[opt]
	elif opt in illegalopts:
	    del options[opt]
	    illegal.append(opt)
	elif opt not in knownopts:
	    del options[opt]
	    unknown.append(opt)
    if dany:
	print >> sys.stderr
    if illegal:
	print >> sys.stderr, "The following options may no more be set "\
	      "in map file"
	print >> sys.stderr, illegal
    for opt in unknown:
	print >> sys.stderr, "WARNING: did not recognize option %s, "\
	      "removed!" % opt

def parse(lines):
    options = {}
    ln = 0
    while ln < len(lines):
	line = lines[ln].split("#")[0].strip() # comments & s/e whitespace
	ln += 1
	if line == "":
	    continue
	option, value = line.split(':', 1)
	option = option.strip().lower()
	value = value.strip()
	if value[:11] == "\multiline:":
	    delim = value[11:].strip() + "\n"
	    value = ""
	    while ln < len(lines) and lines[ln] != delim:
		value += lines[ln]
		ln += 1
	    ln += 1
	options[option] = value
    return options

FILLED = 'x#' # Consider fuel stations as a filled block
REC_UL = 's'
REC_UR = 'a'
REC_DL = 'w'
REC_DR = 'q'
WALL = FILLED + REC_UL + REC_UR + REC_DL + REC_DR
ATTRACT = '$'

BCLICKS = 35 * 64

dirs = ((1, 0), (1, -1), (0, -1), (-1, -1), (-1, 0), (-1, 1), (0, 1), (1, 1))

class Wrapcoords(object):
    def __init__(self, width, height, x = 0, y = 0):
	self.width = width
	self.height = height
	self.x = x % width
	self.y = y % height
    def __eq__(self, other):
	return self.x == other.x and self.y == other.y
    def copy(self):
	return Wrapcoords(self.width, self.height, self.x, self.y)
    def godir(self, dir):
	self.x += dirs[dir][0]
	self.y += dirs[dir][1]
	self.x %= self.width
	self.y %= self.height
    def dist2(self, other):
	x = (self.x - other.x + self.width // 2) % self.width - self.width // 2
	y = (self.y - other.y + self.height//2) % self.height - self.height //2
	return x * x + y * y
# Give class Wrapcoords methods called .u() etc that return a separate
# object one block above etc
dirnames = "r ur u ul l dl d dr".split()
for i in range(8):
    def movedir(self, dir = i):
	result = self.copy()
	result.godir(dir)
	return result
    setattr(Wrapcoords, dirnames[i], movedir)
del i

class Map(object):
    def __init__(self, data, width, height):
	self.data = data
	self.width = width
	self.height = height
    def __getitem__(self, coords):
	return self.data[coords.y][coords.x]
    def __setitem__(self, coords, value):
	self.data[coords.y][coords.x] = value
    def noedge(self, loc, d):
	# Return whether there's solid wall to the right of the
	# direction d. Used when following polygon edge.
	if d == 0 :
	    return self[loc] in FILLED + REC_UL + REC_UR
	elif d == 1:
	    return self[loc.u()] in FILLED + REC_DL + REC_DR
	elif d == 2:
	    return self[loc.u()] in FILLED + REC_UL + REC_DL
	elif d == 3:
	    return self[loc.ul()] in FILLED + REC_UR + REC_DR
	elif d == 4:
	    return self[loc.ul()] in FILLED + REC_DL + REC_DR
	elif d == 5:
	    return self[loc.l()] in FILLED + REC_UR + REC_UL
	elif d == 6:
	    return self[loc.l()] in FILLED + REC_DR + REC_UR
	elif d == 7:
	    return self[loc] in FILLED + REC_DL + REC_UL
    def coords(self):
	for y in range(self.height):
	    for x in range(self.width):
		yield Wrapcoords(self.width, self.height, x, y)
    def ncoords(self):
	return [(x, y) for y in range(self.height) for x in range(self.width)]
    def scratch(self):
	return Map([[None] * self.width for _ in range(self.height)],
		   self.width, self.height)

class Struct:
    pass

def dist2(x1, y1, x2, y2, width, height):
    x = (x1 - x2 + (width >> 1)) % width - (width >> 1)
    y = (y1 - y2 + (height >> 1)) % height - (height >> 1)
    return 1. * x * x + 1. * y * y

def closestteam(loc, bases):
    maxd = 30000 * 30000
    for bs in bases:
	if loc.dist2(bs.loc) < maxd:
	    maxd = loc.dist2(bs.loc)
	    ans = bs.team
    return ans

def poly_partition(map):
    ways = {}
    for i in FILLED:
	ways[i] = [0, 2, 4, 6]
    ways[REC_UR] = [0, 2]
    ways[REC_UL] = [2, 4]
    ways[REC_DL] = [4, 6]
    ways[REC_DR] = [6, 0]
    mapparts = map.scratch()
    partnum = 0
    width, height = map.width, map.height
    for x, y in map.ncoords():
	if map.data[y][x] in WALL and mapparts.data[y][x] is None:
	    left = [(x, y)]
	    mapparts.data[y][x] = (partnum, x, y)
	    while left:
		x2, y2 = left.pop()
		c = map.data[y2 % height][x2 % width]
		for d in ways[c]:
		    x3 = x2 + dirs[d][0]
		    y3 = y2 + dirs[d][1]
		    xw = x3 % width
		    yw = y3 % height
		    if map.data[yw][xw] not in WALL or \
			   ((d + 4) % 8) not in ways[map.data[yw][xw]]:
			continue
		    if mapparts.data[yw][xw] is None:
			left.append((x3, y3))
			mapparts.data[yw][xw] = (partnum, x3, y3)
			continue
	    partnum += 1
    return mapparts, partnum

def poly_findedges(map, mapparts, partcount):
    partpolys = [[] for _ in range(partcount)]
    done = map.scratch()
    # Follow the edge of a polygon and return the polygon.
    def tracepoly(x, y, startdir, okdir = None):
	dir = startdir
	l = Wrapcoords(map.width, map.height, x, y)
	startloc = l.copy()
	poly = []
	while 1:
	    poly.append((x, y, dirs[dir]))
	    x += dirs[dir][0]
	    y += dirs[dir][1]
	    if dir in [6, 7]:
		done[l] = 1
	    elif dir == 5:
		done[l.l()] = 1
	    elif dir == 3:
		done[l.ul()] = 1
	    elif dir == 1:
		done[l.u()] = 1
	    l.godir(dir)
	    dir = (dir + 3) % 8  # always turn as much left as possible
	    while map.noedge(l, dir):
		dir = (dir - 1) % 8  # that much wasn't possible
	    if l == startloc and dir == startdir:
		break
	    if okdir is not None:
		if dir not in okdir:
		    return None
		if dir == 2:
		    done[l.u()] = -1
	return poly

    width, height = map.width, map.height
    for x, y in map.ncoords():
	if done.data[y][x]:
	    continue
	block = map.data[y][x]
	if block not in WALL:
	    continue
	if block == REC_UR:
	    dir = 7
	    dx = dy = 0
	elif block == REC_DR:
	    dx = 1
	    dy = 0
	    dir = 5
	elif block == REC_DL:
	    dx = dy = 1
	    dir = 3
	elif block == REC_UL:
	    dx = 0
	    dy = 1
	    dir = 1
	elif block in FILLED + REC_DL + REC_UL and \
	       map.data[y][(x-1)%width] not in FILLED + REC_DR + REC_UR:
	    dx = dy = 0
	    dir = 6
	else:
	    continue
	sl = mapparts.data[y][x]
	partpolys[sl[0]].append(tracepoly(sl[1] + dx, sl[2] + dy, dir))

    # Handle special (wrapping) edges that aren't detected above
    for loc in map.coords():
	if loc.y > 0:
	    break
	if map[loc] not in FILLED + REC_DL + REC_UL and \
	   map[loc.l()] in FILLED and done[loc] != -1:
	    done[loc] = -1
	    sl = mapparts[loc.l()]
	    poly = tracepoly(sl[1] + 1, sl[2] + 1, 2, (0, 2, 4))
	    if poly:
		partpolys[sl[0]].append(poly)

    loc = Wrapcoords(map.width, map.height, 0, 0)
    while 1:
	if map[loc] in FILLED and map[loc.u()] not in FILLED + REC_DL + REC_DR:
	    sl = mapparts[loc]
	    poly = tracepoly(sl[1] + 1, sl[2], 4, (4,))
	    if poly:
		partpolys[sl[0]].append(poly)
	if map[loc] in FILLED and map[loc.d()] not in FILLED + REC_UR + REC_UL:
	    sl = mapparts[loc]
	    poly = tracepoly(sl[1], sl[2] + 1, 0, (0,))
	    if poly:
		partpolys[sl[0]].append(poly)
	loc.godir(6)
	if loc.y == 0:
	    break
    return partpolys

def polydir(poly):
    xd, yd = poly[-1][2]
    wind = 0
    for point in poly:
	if (yd >= 0) ^ (point[2][1] >= 0):
	    side = xd * point[2][1] - point[2][0] * yd
	    if side != 0:
		if side > 0:
		    wind -= 1
		else:
		    wind += 1
	xd, yd = point[2]
    return wind // 2

def findmin(part1, part2, mindist, distfunc2):
    from math import sqrt
    result = None
    mind = sqrt(mindist)
    far = (mind + 4) ** 2
    swap = 0
    if len(part1) > len(part2):
	part1, part2 = part2, part1
	swap = 1
    for k in range(len(part1)):
	x = part1[k][0]
	y = part1[k][1]
	l = 0
	while l < len(part2):
	    dx, dy = distfunc2(part2[l][0] - x, part2[l][1] - y)
	    d = dx **2 + dy ** 2
	    if d < mindist:
		mindist = d
		mind = sqrt(mindist)
		far = (mind + 4) ** 2
		result = (mindist, k, l, dx, dy)
		l += 1
	    elif d < far:
		l += 1
	    else:
		l += int((sqrt(d) - mind) / 1.4143)
    if not swap or not result:
	return result
    return (result[0], result[2], result[1], -result[3], -result[4])

class Polynode(object):
    def __init__(self, poly):
	self.poly = poly
	self.linkstarts = []
	self.linktargets = []
	self.dist = (1e98, 0, 0)
    def traverse(self, x, y, start, reslist):
	i = start
	while 1:
	    while i in self.linkstarts:
		j = self.linkstarts.index(i)
		lt = self.linktargets[j]
		reslist.append((x, y, (lt.dx, lt.dy), 1))
		lt.targetnode.traverse(x + lt.dx, y + lt.dy,
				       lt.targetindex, reslist)
		reslist.append((x + lt.dx, y + lt.dy, (-lt.dx, -lt.dy), 1))
		del self.linkstarts[j]
		del self.linktargets[j]
	    dx, dy = self.poly[i][2][0], self.poly[i][2][1]
	    reslist.append((x, y, (dx, dy), 0))
	    x += dx
	    y += dy
	    i = (i + 1) % len(self.poly)
	    if i == start:
		break

def poly_handle_totalwrap(poly, map):
    res = []
    prev = (0, 0)
    for r in poly:
	if max(prev[0], prev[1]) > 1:
	    for _ in range(max(prev[2][0], prev[2][1]) - 1):
		# Don't let findmin skip optimization break
		res.append((r[0], r[1], (0, 0), r[3]))
	res.append(r)
	prev = r
    def minwrap(dx, dy, wx = map.width, wy = map.height):
	dx2 = (dx + (wx >> 1)) % wx - (wx >> 1)
	dy2 = (dy + (wy >> 1)) % wy - (wy >> 1)
	if dx2 == dx and dy2 == dy:
	    if abs(dx2) > abs(dy2):
		if dx2 > 0:
		    dx2 -= wx
		else:
		    dx2 += wx
	    else:
		if dy2 > 0:
		    dy2 -= wy
		else:
		    dy2 += wy
	return dx2, dy2
    r = findmin(res, res, 1e99, minwrap)
    wx = res[r[2]][0] - res[r[1]][0] - r[3]
    wy = res[r[2]][1] - res[r[1]][1] - r[4]
    assert wx % map.width == 0
    assert wy % map.height == 0
    a = (1, 0, wx // map.width)
    b = (0, 1, wy // map.height)
    while a[2]:
	d = b[2] // a[2]
	a, b = (b[0] - d * a[0], b[1] - d * a[1], b[2] - d * a[2]), a
    assert abs(b[2]) == 1
    wx2 = b[1] * b[2]
    wy2 = -b[0] * b[2]
    assert wy // map.height * wx2 + (-wx // map.width) * wy2 == 1
    if r[1] < r[2]:
	p1 = res[r[1]:r[2] + 1]
	p2 = res[r[2]:] + res[:r[1] + 1]
    else:
	p1 = res[r[1]:] + res[:r[2] + 1]
	p2 = res[r[2]:r[1] + 1]
    p1[-1] = (p1[-1][0], p1[-1][1], (-r[3], -r[4]), 1)
    p2[-1] = (p2[-1][0], p2[-1][1], (r[3], r[4]), 1)
    wx2 *= map.width
    wy2 *= map.height
    p2 = [(p[0] + wx2, p[1] + wy2, p[2], p[3]) for p in p2]
    def minwrap(dx, dy, wx = wx, wy = wy, l2 = wx **2 + wy ** 2):
	d = int(round((wx * dx + wy * dy) / l2))
	return dx - d * wx, dy - d * wy
    r2 = findmin(p1, p2, 1e99, minwrap)
    assert (p1[r2[1]][0] + r2[3] - p2[r2[2]][0]) % map.width == 0
    assert (p1[r2[1]][1] + r2[4] - p2[r2[2]][1]) % map.height == 0
    res = p1[:r2[1]]
    res.append((p1[r2[1]][0], p1[r2[1]][1], (r2[3], r2[4]), 1))
    res += p2[r2[2]:] + p2[:r2[2]]
    res.append((p2[r2[2]][0], p2[r2[2]][1], (-r2[3], -r2[4]), 1))
    res += p1[r2[1]:]
    x, y = res[0][0], res[0][1]
    result = []
    for p in res:
	result.append((x, y, p[2], p[3]))
	x += p[2][0]
	y += p[2][1]
    return result

def poly_link(polys, map):
    neglist = []
    poslist = []
    zerolist = []
    for p in polys:
	direction = polydir(p)
	if direction > 0:
	    poslist.append(p)
	elif direction < 0:
	    neglist.append(p)
	else:
	    zerolist.append(p)
    if poslist and not zerolist:
	assert len(poslist) <= 1, "Can't have multiple + polygons in one part"
	def minwrap(dx, dy):
	    return dx, dy
	extrawrap = 0
    elif not zerolist and not poslist:
	def minwrap(dx, dy, wx = map.width, wy = map.height):
	    dx = (dx + (wx >> 1)) % wx - (wx >> 1)
	    dy = (dy + (wy >> 1)) % wy - (wy >> 1)
	    return dx, dy
	extrawrap = 1
    else:
	assert not poslist, "Can't have both zero and + polygons in one part"
	assert len(zerolist) == 2, "BUG"
	dx = dy = 0
	for p in zerolist[0]:
	    dx += p[2][0]
	    dy += p[2][1]
	assert abs(dx) + abs(dy) > 0, "Wrapping polygon with no wrap?"
	assert dx % map.width == 0, "Garbled coordinates"
	assert dy % map.height == 0, "Garbled coordinates"
	def minwrap(dx, dy, wx = dx, wy = dy, l2 = dx **2 + dy ** 2):
	    d = int(round((wx * dx + wy * dy) / l2))
	    return dx - d * wx, dy - d * wy
	extrawrap = 0
    totlist = [Polynode(p) for p in neglist + poslist + zerolist]
    res = last = totlist.pop()

    while totlist:
	for j in totlist:
	    result = findmin(last.poly, j.poly, j.dist[0], minwrap)
	    if result:
		j.dist = (result[0], result[1], result[2], result[3],
			  result[4], last)
	mindist = 1e99
	mink = -2.1
	for j in totlist:
	    if j.dist[0] < mindist:
		mink = j
		mindist = j.dist[0]
	totlist.remove(mink)
	r = mink.dist[5]
	r.linkstarts.append(mink.dist[1])
	s = Struct()
	(s.targetnode, s.targetindex, s.dx, s.dy) = (mink, mink.dist[2],
						    mink.dist[3], mink.dist[4])
	r.linktargets.append(s)
	last = mink
    result = []
    res.traverse(res.poly[0][0], res.poly[0][1], 0, result)
    if extrawrap:
	result = poly_handle_totalwrap(result, map)
    return result

def makepolys(map):
    print >> sys.stderr, "   Partitioning map...",
    mapparts, partcount = poly_partition(map)
    print >> sys.stderr, partcount, "parts."
    print >> sys.stderr, "   Finding edges...",
    partpolys = poly_findedges(map, mapparts, partcount)
    print >> sys.stderr, "done."
    allpolys = []
    print >> sys.stderr, "   Linking each part...",
    allpolys = [poly_link(polys, map) for polys in partpolys]
    print >> sys.stderr, "done."
    return allpolys

def convert(options):
    height = int(options['mapheight'])
    width = int(options['mapwidth'])
    map = options['mapdata'].splitlines()
    if options.get('edgewrap') not in ['yes', 'on', 'true']: #default off
	height += 2
	width += 2
	map = [' ' + line + ' ' for line in map]
	map = [' ' * width] + map + [' ' * width]

	options['mapdata'] = "\n".join(map)+'\n'
    options['mapwidth'] = `width * 35`
    options['mapheight'] = `height * 35`
    map = Map(map, width, height)

    print >> sys.stderr, "Finding special map features...",
    # First, find the locations of different map features other than walls
    # and put them in these lists.
    asteroidconcs = []
    bases = []
    balls = []
    cannons = []
    checks = [None] * 27  # 1 extra so it always ends with None
    decors = []
    emptytreasures = []
    fuels = []
    frictions = []
    gravs = []
    itemconcs = []
    targets = []
    wormholes = []
    ASTEROIDCONC = '&'
    BASES = '_0123456789'
    BALL = '*'
    CANNONS = 'rdfc'
    CHECKPOINTS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    DECOR = 'bhgyt'
    EMPTYTREASURE = '^'
    FUEL = '#'
    FRICTION = 'z'
    GRAVS = '+-><imkj'
    ITEMCONC = '%'
    SPACE = ' '
    TARGET = '!'
    WALLS = 'xsawq'       # WALL includes fuel block
    WORMHOLES = '@()'
    ALL = ASTEROIDCONC + BASES + BALL + CHECKPOINTS + CANNONS + DECOR + EMPTYTREASURE + FUEL + FRICTION + GRAVS + ITEMCONC + TARGET + WORMHOLES
    IGNORE = ATTRACT + SPACE + WALLS
    for x, y in map.ncoords():
	block = map.data[y][x]
	if block in IGNORE:
	    continue
	if block not in ALL:
	    print >>sys.stderr, "WARNING: Unknown block: '%s'"    % block
	    continue
	loc = Wrapcoords(map.width, map.height, x, y)
	thing = Struct()
	thing.x = loc.x * BCLICKS + BCLICKS // 2
	thing.y = (height - loc.y - 1) % height * BCLICKS + BCLICKS // 2
	thing.loc = loc
	thing.team = -1
	thing.type = ''
	if block == ASTEROIDCONC:
	    asteroidconcs.append(thing)
	elif block in BASES:
	    if map[loc] == '_':
		thing.team = 0
	    else:
		thing.team = ord(map[loc]) - ord('0')
	    thing.dir = 32
	    if map[loc.u()] in ATTRACT:
		thing.dir = 32
	    elif map[loc.d()] in ATTRACT:
		thing.dir = 96
	    elif map[loc.r()] in ATTRACT:
		thing.dir = 0
	    elif map[loc.l()] in ATTRACT:
		thing.dir = 64
	    thing.loc = loc
	    bases.append(thing)
	elif block == BALL:
	    balls.append(thing)
	elif block in CANNONS:
	    thing.dir = 'frdc'.index(block) * 32
	    cannons.append(thing)
	elif block in CHECKPOINTS:
	    checks[ord(map[loc]) - ord('A')] = thing
	elif block in DECOR:
	    print >>sys.stderr, "WARNING: can't yet convert decor."
	    continue
	elif block == EMPTYTREASURE:
	    print >>sys.stderr, "WARNING: can't yet convert empty treasure."
	    continue
	elif block == FUEL:
	    fuels.append(thing)
	elif block == FRICTION:
	    print >>sys.stderr, "WARNING: can't yet convert friction block."
	    continue
	elif block in GRAVS:
	    print >>sys.stderr, "WARNING: can't yet convert gravs."
	    continue
	elif block == ITEMCONC:
	    itemconcs.append(thing)
	elif block == TARGET:
	    targets.append(thing)
	elif block in WORMHOLES:
	    if map[loc] == '@':
		thing.type = 'normal'
	    elif map[loc] == '(':
		thing.type = 'in'
	    elif map[loc] == ')':
		thing.type = 'out'
	    wormholes.append(thing)
    if not bases:
	print >>sys.stderr, "Map has no bases???"
	sys.exit(1)
    if KEEP_MAPDATA:
	def sort_for_old(things):
	    things[:] = [(thing.x, thing.y, thing) for thing in things]
	    things.sort()
	    things[:] = [thing[-1] for thing in things]
	sort_for_old(bases)
	sort_for_old(fuels)
	sort_for_old(cannons)
    # Balls belong to the team that has the nearest base
    if options.get('teamplay'):
	for ball in balls:
	    ball.team = closestteam(ball.loc, bases)
	for target in targets:
	    target.team = closestteam(target.loc, bases)
    # In race mode, bases are ordered according to the distance from the
    # first checkpoint. Confuses order for old clients (above)
    if (options.get('timing') or options.get('race')) in ['yes', 'on', 'true']:
	bases = [(checks[0].loc.dist2(b.loc), b) for b in bases]
	bases.sort()
	bases = [b[1] for b in bases]
    # If teamfuel is on, the fuel belongs to the team with the nearest base
    if options.get('teamfuel') in ['yes', 'on', 'true']: #default off
	for fuel in fuels:
	    fuel.team = closestteam(fuel.loc, bases)
    if options.get('teamcannons') in ['yes', 'on', 'true']: #default off
	for cannon in cannons:
	    cannon.team = closestteam(cannon.loc, bases)

    print >> sys.stderr, "done."
    print >> sys.stderr, "Creating polygons:"
    polys = makepolys(map)

    # Turn this on if you want to randomize the map edges
    # Needs some changes to work now because of other changes elsewhere
    if 0:
	from math import sin
	for p in polys:
	    for l in p:
		l[0] = int(l[0] + sin(1. * l[0] * l[1] / 2786) * (2240 / 3)) % mxc
		l[1] = int(l[1] + sin(1. * l[0] * l[1] / 1523) * (2240 / 3)) % myc

    print >> sys.stderr, "Writing converted map...",
    print '<XPilotMap version="1.2">'

    if not KEEP_MAPDATA:
	del options['mapdata']

    print '<GeneralOptions>'

    def xmlencode(text):
	repl = (('&', '&amp;'), ('"', '&quot;'), ("'", '&apos;'),
		('<', '&lt;'), ('>', '&gt;'), ('\n', '&#xA;'))
	for t, r in repl:
	    text = text.replace(t, r)
	return text
    for name, value in options.items():
	print '<Option name="%s" value="%s"/>' % (name, xmlencode(value))
    print '</GeneralOptions>'
    print '<Edgestyle id="xpbluehidden" width="-1" color="4E7CFF" style="0"/>'
    print '<Edgestyle id="xpredhidden" width="-1" color="FF3A27" style="0"/>'
    print '<Edgestyle id="yellow" width="2" color="FFFF00" style="0"/>'
    print '<Edgestyle id="white" width="2" color="FFFFFF" style="0"/>'
    print '<Edgestyle id="invisiedge" width="-1" color="0" style="0"/>'
    print '<Polystyle id="xpblue" color="4E7CFF" defedge="xpbluehidden" flags="1"/>'
    print '<Polystyle id="xpred" color="FF3A27" defedge="xpredhidden" flags="1"/>'
    print '<Polystyle id="emptyyellow" color="FF" defedge="yellow" flags="0"/>'
    print '<Polystyle id="emptywhite" color="FF" defedge="white" flags="0"/>'
    print '<Polystyle id="invisible" color="0" defedge="invisiedge" flags="12"/>'

    def printedge(dx, dy, prevh, curh):
	if curh and not prevh:
	    sstr = ' style="internal"'
	elif prevh and not curh:
	    sstr = ' style="xpbluehidden"'
	else:
	    sstr = ''
	print '<Offset x="%d" y="%d"%s/>' % (dx, dy, sstr)

    for p in polys:
	while p[-1][2][0] * p[0][2][1] == p[-1][2][1] * p[0][2][0]:
	    p.insert(0, p.pop())
	edges = [(point[2][0] * BCLICKS, -point[2][1]* BCLICKS, point[3])
		 for point in p if point[2] != (0, 0)]
	x = (p[0][0] % map.width) * BCLICKS
	y = (-p[0][1] % map.height) * BCLICKS

	if WALL_SHRINK_HACK:
	    s = WALL_SHRINK_HACK
	    ch = [(-edge[1] // BCLICKS, edge[0] // BCLICKS) for edge in edges]
	    ch2 = []
	    for ((x1, y1), (x2, y2)) in zip(ch[-1:] + ch[:-1], ch):
		if x1 * x2 >= 0 and y1 * y2 >= 0:
		    ch2.append((x1 or x2, y1 or y2))
		elif x1 and x2 and y1 and y2:
		    ch2.append((x1 + x2, y1 + y2))
		else:
		    ch2.append(((x1 + x2) * 2 + x1 * (not y1) + x2 * (not y2
			), (y1 + y2) * 2 + y1 * (not x1) + y2 * (not x2)))
	    ch2 = [(x1 * s, y1 * s) for (x1, y1) in ch2]
	    x = (x + ch2[0][0]) % (map.width * BCLICKS)
	    y = (y + ch2[0][1]) % (map.height * BCLICKS)
	    ch2 = [(x2 - x1, y2 - y1) for ((x1,y1), (x2,y2)) in zip(ch2,
		ch2[1:] + ch2[:1])]
	    edges = [(edge[0] + ch[0], edge[1] + ch[1], edge[2]) for edge, ch
			in zip(edges, ch2)]
	print '<Polygon x="%d" y="%d" style="xpblue">' % (x, y)
	prevh = 0
	curh = edges[0][2]
	curx = cury = 0
	sstr = ''
	for edge in edges:
	    (dx, dy, h) = edge
	    if dx * cury != dy * curx or curh != h:
		printedge(curx, cury, prevh, curh)
		curx = dx
		cury = dy
		prevh = curh
		curh = h
	    else:
		curx += dx
		cury += dy
	printedge(curx, cury, prevh, curh)
	print "</Polygon>"
# The styles of these polygons will be changed later...
    for aconc in asteroidconcs:
	print '<AsteroidConcentrator x="%d" y="%d"/>' % (aconc.x, aconc.y)
    for ball in balls:
	print '<Ball team="%d" x="%d" y="%d"/>' % (ball.team, ball.x, ball.y - 479)
	print '<BallArea>'
	print '<Polygon x="%d" y="%d" style="xpblue">' % (ball.x - 1120, ball.y - 1120)
	print '<Offset x="2240" y="0"/> <Offset x="0" y="2240"/>'
	print '<Offset x="-2240" y="0"/> <Offset x="0" y="-2240"/>'
	print '</Polygon></BallArea>'
	print '<BallTarget team="%d">' % ball.team
	print '<Polygon x="%d" y="%d" style="xpred">' % (ball.x - 480, ball.y - 480)
	print '<Offset x="960" y="0"/> <Offset x="0" y="960"/>'
	print '<Offset x="-960" y="0"/> <Offset x="0" y="-960"/>'
	print '</Polygon></BallTarget>'
    for base in bases:
	print '<Base team="%d" x="%d" y="%d" dir = "%d"/>' % \
	      (base.team, base.x, base.y, base.dir)
    for fuel in fuels:
	print ('<Fuel x="%d" y="%d"' % (fuel.x, fuel.y)),
	sys.stdout.softspace = 0
	if fuel.team != -1:
	    print (' team="%d"' % fuel.team),
	print '/>'
    for target in targets:
	text = '<Target x="%d" y="%d"' % (target.x, target.y)
	if target.team != -1:
	    text += ' team="%d"' % target.team
	text += '>'
	print text
	print '<Polygon x="%d" y="%d" style="xpred">' % (target.x - 1120, target.y - 1120)
	print '<Style state="destroyed" id="invisible"/>'
	print '<Offset x="2240" y="0"/> <Offset x="0" y="2240"/>'
	print '<Offset x="-2240" y="0"/> <Offset x="0" y="-2240"/>'
	print '</Polygon></Target>'
    for check in checks:
	if not check:
	    break
	print '<Check x="%d" y="%d"/>' % (check.x, check.y)
    for cannon in cannons:
	offsets = ((-373, 0), (-747, 1120), (0, -2240), (747, 1120))
	for _ in range(cannon.dir // 32):
	    offsets = [(-y, x) for x, y in offsets]
	cannon.x += offsets[0][0]
	cannon.y += offsets[0][1]
	print ('<Cannon x="%d" y="%d" dir="%d"' % \
	       (cannon.x, cannon.y, cannon.dir)),
	sys.stdout.softspace = 0
	if cannon.team != -1:
	    print (' team="%d"' % cannon.team),
	print '>\n<Polygon x="%d" y="%d" style="emptywhite">' % \
	      (cannon.x, cannon.y)
	print '<Style state="destroyed" id="invisible"/>'
	for x, y in offsets[1:]:
	    print '<Offset x="%d" y="%d"/>' % (x, y)
	print '</Polygon></Cannon>'
    for iconc in itemconcs:
	print '<ItemConcentrator x="%d" y="%d"/>' % (iconc.x, iconc.y)
    for wormhole in wormholes:
	print '<Wormhole x="%d" y="%d" type="%s">' % \
		(wormhole.x, wormhole.y, wormhole.type)
	if wormhole.type != 'out':
	    print '<Polygon x="%d" y="%d" style="xpred">' % \
		(wormhole.x - 1120, wormhole.y - 1120)
	    print '<Offset x="2240" y="0"/> <Offset x="0" y="2240"/>'
	    print '<Offset x="-2240" y="0"/> <Offset x="0" y="-2240"/>'
	    print '</Polygon>'
	print '</Wormhole>'
    print "</XPilotMap>"
    print >> sys.stderr, "done."

if __name__ == '__main__':
    import sys
    if len(sys.argv) != 2:
	print "Give 1 argument, the map file name"
	sys.exit(1)
    file = open(sys.argv[1])
    lines = file.readlines()
    file.close()
    options = parse(lines)
    checkopts(options)
    convert(options)



import gdkpixbuf
import gtk
# The gdkpixbuf module doesn't seem too reliable, for example the
# render_pixmap_and_mask method seemed to always cause a segmentation fault
# if the gtk module had not been imported...
from javastuff import *
from MapObjectPopup import MapObjectPopup
import EditorDialog
import copy

class MapObject:
    def __init__(self, x, y, width, height, img):
        self.bounds = Rectangle(x - width / 2, y - height / 2, width, height)
        self.x = x
        self.y = y
        self.img = img

    def getBounds(self):
        return self.bounds

    def setBounds(self, bounds):
        self.bounds = bounds

    def contains(self, canvas, p):
        return len(canvas.wrapOffsets(self.getBounds(), p))

    def getZOrder(self):
        return 10

    def getCreateHandler(self):
        return CreateHandler(self)

    def moveTo(self, x, y):
        self.x = x
        self.y = y
        self.bounds.x = x - self.bounds.width / 2
        self.bounds.y = y - self.bounds.height / 2

    def getPreviewShape(self):
        return self.getBounds()

    def paint(self, di):
        if self.img is None:
            return
        r = self.getBounds()
        tx = di.tx
        pixmap = self.img[di.scale * 64]
        p = Point(r.x, r.y)
        p = tx(p)
        gtk.draw_pixmap(di.area, di.gc, pixmap, 0, 0, p.x, p.y - pixmap.height,
                            pixmap.width, pixmap.height)

    def checkEvent(self, canvas, me):
        if self.contains(canvas, me.point):
            if me.button == 1:
                if canvas.isErase():
                    canvas.getModel().removeObject(self)
                    canvas.repaint()
                    canvas.saveUndo()
                else:
                    canvas.setCanvasEventHandler(MoveHandler(self, me))
                return 1
            elif me.button == 3:
                MapObjectPopup(self, me.orig, canvas)
                return 1
        return 0

    def getPropertyEditor(self, canvas):
        class C(gtk.GtkLabel):
            def apply(self):
                return 1
        editor = C("No editable properties")
        editor.buttons = EditorDialog.CLOSE
        editor.show()
        return editor


class CreateHandler(MouseEventHandler):
    def __init__(self, object):
        self.object = object
        self.preview = object.getPreviewShape()
        self.toBeRemoved = None

    def mouseMoved(self, evt):
        c = evt.canvas
        gc = c.previewgc
        if self.toBeRemoved is not None:
            c.drawShape(self.toBeRemoved, gc)
        s = copy.copy(self.preview)
        s.transform(AffineTransform(evt.x - self.object.x,
                                    evt.y - self.object.y))
        c.drawShape(s)
        self.toBeRemoved = s

    def mousePressed(self, evt):
        c = evt.canvas
        c.getModel().addToFront(self.object)
        mapsize = c.model.options.size
        self.object.moveTo(evt.x % mapsize.width, evt.y % mapsize.height)
        c.setCanvasEventHandler(None)
        c.repaint()
        c.saveUndo()

class MoveHandler(MouseEventHandler):
    def __init__(self, obj, evt):
        self.object = obj
        self.offset = Point(evt.x - obj.x, evt.y - obj.y)
        self.preview = obj.getPreviewShape()
        self.toBeRemoved = None

    def mouseMoved(self, evt):
        c = evt.canvas
        if self.toBeRemoved is not None:
            c.drawShape(self.toBeRemoved)
        s = copy.copy(self.preview)
        s.transform(AffineTransform(evt.x - self.offset.x - self.object.x,
                                    evt.y - self.offset.y - self.object.y))
        c.drawShape(s)
        self.toBeRemoved = s

    def mouseReleased(self, evt):
        c = evt.canvas
        mapsize = c.model.options.size
        self.object.moveTo((evt.x - self.offset.x) % mapsize.width,
                           (evt.y - self.offset.y) % mapsize.height)
        c.setCanvasEventHandler(None)
        c.repaint()
        c.saveUndo()



from MapObject import MapObject


class MapCheckPoint(MapObject):
    def __init__(self, x = 0, y = 0, bmstore = None):
        MapObject.__init__(self, x, y, 35 * 64, 35 * 64,
                           bmstore["checkpoint.gif"])

    def printXml(self, file):
        file.write('<Check x="%d" y="%d"/>\n' % (self.x, self.y))
